from pyadlml.constants import *
from pyadlml.dataset import load_act_assist
import unittest
import sys
import pathlib
working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))


class TestDirtyDataset(unittest.TestCase):
    def setUp(self):
        dataset_dir = str(script_directory) + '/datasets/diry_dataset'
        self.data = load_act_assist(dataset_dir)
        self.df_acts = self.data['activities']
        self.df_devs = self.data['devices']

    def test_stats_activities(self):
        from pyadlml.dataset.stats.activities import activities_dist, activities_count, \
            activities_transitions, activities_duration_dist

        df_acts = self.df_acts.copy()
        lst = self.data['list_activities']

        act_count = activities_count(df_acts, lst)
        assert len(act_count) == len(lst)
        act_count = activities_count(df_acts)
        assert len(act_count) == 2

        act_trans = activities_transitions(df_acts, lst)
        assert len(act_trans) == len(lst)
        assert act_trans.values.sum() == len(df_acts) - 1
        act_trans = activities_transitions(df_acts)
        assert len(act_trans) == 2

        act_durs = activities_duration_dist(df_acts, lst)
        assert len(act_durs) == len(lst)
        act_durs = activities_duration_dist(df_acts)
        assert len(act_durs) == 2

        act_dist = activities_dist(df_acts, lst, n=100)
        assert len(act_dist.columns) == len(lst)
        assert len(act_dist) == 100
        act_dist = activities_dist(df_acts, n=100)
        assert len(act_dist.columns) == 2
        assert len(act_dist) == 100

    def test_stats_devices(self):
        from pyadlml.dataset.stats.devices import event_count, events_one_day,\
            inter_event_intervals, state_fractions, state_cross_correlation

        df_devs = self.df_devs.copy()
        lst = self.data['device_lst']

        num_rec_devs = 5  # only those devices that appear in devices.csv
        num_rec_binary_devs = 3

        # tcorr = device_tcorr(df)
        tcorr = device_tcorr(df_devs, lst)
        assert tcorr.shape[0] == len(lst) and tcorr.shape[1] == len(lst)

        trigg_c = devices_trigger_count(df_devs)
        assert len(trigg_c) == num_rec_devs
        trigg_c = devices_trigger_count(df_devs, lst)
        assert len(trigg_c) == len(lst)

        trigg_od = device_triggers_one_day(df_devs)
        assert len(trigg_od.columns) == num_rec_devs
        trigg_od = device_triggers_one_day(df_devs,  lst)
        assert len(trigg_od.columns) == len(lst)

        trigg_td = trigger_time_diff(df_devs)
        assert len(trigg_td) == len(df_devs) - 1

        onoff = devices_on_off_stats(df_devs)
        assert len(onoff) == num_rec_binary_devs
        onoff = devices_on_off_stats(df_devs, lst)
        assert len(onoff) == len(lst)

        dc = duration_correlation(df_devs)
        assert len(dc) == num_rec_binary_devs
        dc = duration_correlation(df_devs, lst)
        assert len(dc) == len(lst)

    def test_plot_activities(self):
        from pyadlml.dataset.plot.activities import activities_duration_dist, boxplot_duration, \
            heatmap_transitions, hist_cum_duration, ridge_line, hist_counts

        df = self.data.df_activities_admin
        lst = self.data.lst_activities

        hist_counts(df)
        hist_counts(df, lst_act=lst)

        boxplot_duration(df)
        boxplot_duration(df, lst_act=lst)

        hist_cum_duration(df)
        hist_cum_duration(df, act_lst=lst)

        heatmap_transitions(df)
        heatmap_transitions(df, lst_act=lst)

        ridge_line(df)
        ridge_line(df, lst_act=lst, n=10)

    def test_plot_devices(self):
        from pyadlml.dataset.plot.devices import hist_counts, heatmap_trigger_one_day, heatmap_trigger_time,\
            hist_on_off, boxplot_on_duration, heatmap_cross_correlation, hist_trigger_time_diff

        df = self.data.df_devices
        lst = self.data.lst_devices

        hist_counts(df)
        hist_counts(df_dev=df, lst_dev=lst)

        heatmap_trigger_one_day(df)
        heatmap_trigger_one_day(df, lst_dev=lst)

        heatmap_trigger_time(df)
        heatmap_trigger_time(df, lst_dev=lst)

        hist_trigger_time_diff(df)

        hist_on_off(df)
        hist_on_off(df, lst_dev=lst)

        boxplot_on_duration(df)
        boxplot_on_duration(df, lst_dev=lst)

        heatmap_cross_correlation(df)
        heatmap_cross_correlation(df, lst_dev=lst)


if __name__ == '__main__':
    unittest.main()



import torch
import numpy as np
import pandas as pd
from sklearn.metrics import confusion_matrix
import plotly.graph_objects as go

from pyadlml.metrics import online_confusion_matrix


def plotly_confusion_matrix(y_pred, y_true, labels, per_step=False, scale='linear') -> go.Figure:
    """
    """
    cbarlabel = 'counts' if scale == 'linear' else 'log counts'
    if not per_step:
        if isinstance(y_pred, torch.Tensor):
            y_pred = y_pred.detach().cpu().numpy()

        if isinstance(y_true, torch.Tensor):
            y_true = y_true.detach().cpu().numpy()

        # Row = true, column = prediction
        cm = confusion_matrix(y_true=y_true, y_pred=y_pred,
                              labels=labels).astype(int)
        z = np.log(cm) if scale == 'log' else cm

        fig = go.Figure(
            data=go.Heatmap(
                z=z.T,
                x=labels,
                y=labels,
                customdata=cm.T,
                colorscale='Viridis',
                hovertemplate='Truth: %{y}<br>Pred: %{x}<br>Times: %{customdata} <extra></extra>',
                hoverongaps=False),
        )
        fig.update_layout(
            title="Confusion Matrix",
            xaxis=dict(title="Predicted"),
            yaxis=dict(title="True"),
        )

        return fig
    else:

        nr_steps = len(y_pred)

        if isinstance(y_pred[0], torch.Tensor):
            y_pred = [y.detach().cpu().numpy() for y in y_pred]

        if isinstance(y_true[0], torch.Tensor):
            y_true = [y.detach().cpu().numpy() for y in y_true]

        cms = [confusion_matrix(y_true=yt, y_pred=yp).astype(int)
               for (yp, yt) in zip(y_pred, y_true)]

        fig = go.Figure()
        import plotly.express as px

        for step in range(nr_steps):
            fig.add_trace(
                go.Heatmap(
                    name=f'CM_{step}',
                    z=cms[step],
                    x=labels,
                    y=labels,
                    colorscale='Viridis',
                    hovertemplate='Pred: %{x}<br>Truth: %{y}<br>Times: %{z} <extra></extra>',
                    hoverongaps=False
                ),
            )

        # Make last confusion matrix visible
        fig.data[nr_steps-1].visible = True

        steps = []
        for i in range(len(fig.data)):
            step = dict(
                method="update",
                args=[{"visible": [False] * len(fig.data)},
                      {"title": "Confusion matrix: " + str(i)}],  # layout attribute
            )
            # Toggle i'th trace to "visible"
            step["args"][0]["visible"][i] = True
            steps.append(step)

        sliders = [dict(
            active=10,
            currentvalue={"prefix": "Epoch: "},
            pad={"t": 50},
            steps=steps
        )]

        fig.update_layout(
            sliders=sliders
        )

        # Share colorscale
        for i in range(nr_steps):
            fig.data[i]['coloraxis'] = 'coloraxis'

    return fig


def plotly_online_confusion_matrix(y_true=None, y_pred=None, y_times=None, labels=None, conf_mat=None, scale='linear') -> go.Figure:
    """
    """
    cbarlabel = 'counts' if scale == 'linear' else 'log seconds'
    if conf_mat is None:
        if isinstance(y_pred, torch.Tensor):
            y_pred = y_pred.detach().cpu().numpy()

        if isinstance(y_true, torch.Tensor):
            y_true = y_true.detach().cpu().numpy()

        # Row = true, column = prediction
        cm = online_confusion_matrix(
            y_true, y_pred, y_times, n_classes=len(labels))
    else:
        cm = conf_mat
        labels = conf_mat.index
    z = cm.astype('timedelta64[ns]')/np.timedelta64(1, 'ns')
    z = z if scale == 'linear' else np.log(z)
    z = z.T

    # Create hoverdata, the duration as strings in (D,A,0) and full X names in (D,A,1)
    cd = cm.astype(str).values.T
    # tmp = np.tile(vals.index, (len(vals.columns), 1))
    # cd = np.array([tmp, vals.values.T])
    # d = np.moveaxis(cd, 0, -1)

    fig = go.Figure(
        data=go.Heatmap(
            z=z,
            x=labels,
            y=labels,
            customdata=cd,
            colorscale='Viridis',
            hovertemplate='Truth: %{y}<br>Pred: %{x}<br>Overlap: %{customdata} <extra></extra>',
            hoverongaps=False),
    )
    fig.update_layout(
        title="Confusion Matrix",
        xaxis=dict(title="Predicted"),
        yaxis=dict(title="True"),
    )

    return fig



import os
import multiprocessing
import numpy as np

ENV_PARALLEL = 'PYADLML_PARALLEL'


def set_parallel(val):
    """ tells code to run execute tasks with dask in parallel whenever possible
    """
    assert isinstance(val, bool)
    os.environ[ENV_PARALLEL] = str(val)
    assert get_parallel() == val


def get_parallel():
    val = os.environ.get(ENV_PARALLEL)
    assert val in ["True", "False"]
    return val == "True"


def get_npartitions():
    """ returns the num of parallel threads that can work on a computer
    """
    return 4*multiprocessing.cpu_count()


def _save_divide(num: np.ndarray, denom: np.ndarray) -> np.ndarray:
    """Safe division, by preventing division by zero.

    Additionally casts to float if input is not already to secure backwards compatibility.
    """
    num = num if isinstance(num, np.floating) else num.astype(np.float64)
    denom = denom if isinstance(
        denom, np.floating) else denom.astype(np.float64)
    denom[denom == 0.0] = 1e-10

    return num / denom


ENV_PRIMARY_COLOR = 'PYADLML_PRIMARY_COLOR'
ENV_SECONDARY_COLOR = 'PYADLML_SECONDARY_COLOR'
ENV_DIV_COLORMAP = 'PYADLML_DIVERGING_COLOR'
ENV_SEQ_COLORMAP = 'PYADLML_SEQUENTIAL_COLOR'
# theming


def set_primary_color(val):
    os.environ[ENV_PRIMARY_COLOR] = val


def set_secondary_color(val):
    os.environ[ENV_SECONDARY_COLOR] = val


def set_diverging_color(val):
    os.environ[ENV_DIV_COLORMAP] = val


def set_sequential_color(val):
    os.environ[ENV_SEQ_COLORMAP] = val


def get_primary_color():
    return os.environ[ENV_PRIMARY_COLOR]


def get_secondary_color():
    return os.environ[ENV_SECONDARY_COLOR]


def get_diverging_color():
    return os.environ[ENV_DIV_COLORMAP]


def get_sequential_color():
    return os.environ[ENV_SEQ_COLORMAP]



import numpy as np
import pandas as pd
from pyadlml.constants import ACTIVITY, END_TIME, OTHER, START_TIME, TIME, OTHER_MIN_DIFF
from pyadlml.dataset._core.activities import is_activity_df
from pyadlml.util import _save_divide


"""
metrics oriented at torch.metrics
fp, 
/home/chris/venvs/ma/lib/python3.8/site-packages/torchmetrics/functional/classification/stat_scores.py

"""


def online_true_positive_rate(y_true: np.ndarray, y_pred: np.ndarray, times: np.ndarray, n_classes: int, average: str = None):
    """ Calculate the true positive rate/recall.

    Parameters
    ----------

    average: str one of ['micro', 'macro', 'weighted']   

    """
    assert average in [None, 'micro', 'macro', 'weighted']
    df_confmat = online_confusion_matrix(y_true, y_pred, times, n_classes)

    if average == 'micro':
        cm = df_confmat.values
        tp = np.diag(cm)
        fn = cm.sum(axis=0) - tp

        FN, TP = fn.sum(), tp.sum()
        return TP/(TP + FN)

    elif average == 'macro':
        cm = df_confmat.values
        tp = np.diag(cm)
        fn = cm.sum(axis=0) - tp
        score = _save_divide(tp, (tp + fn))

        return (score*(1/n_classes)).sum(-1)


def online_positive_predictive_value(y_true: np.ndarray, y_pred: np.ndarray, times: np.ndarray, n_classes: int, average: str = None):
    """ Calculate the positive predictive value/precision.

    Parameters
    ----------

    average: str one of ['micro', 'macro', 'weighted']   
    """
    assert average in [None, 'micro', 'macro', 'weighted']

    # Extract true as columns and preds as rows in nanosecond resolution
    # c_ij = true class i when predicted class is j
    df_confmat = online_confusion_matrix(y_true, y_pred, times, n_classes)

    if average == 'micro':
        cm = df_confmat.values
        tp = np.diag(cm)
        fp = cm.sum(axis=1) - tp

        FP, TP = fp.sum(), tp.sum()
        return TP/(TP + FP)

    elif average == 'macro':
        cm = df_confmat.values
        tp = np.diag(cm)
        fp = cm.sum(axis=1) - tp
        score = tp / (tp + fp)

        return (score*(1/n_classes)).sum(-1)


def online_accuracy(y_true: np.ndarray, y_pred: np.ndarray, times: np.ndarray, n_classes: int, average: str = None):
    """ Calculate the overlap 

    Parameters
    ----------

    average: str one of ['micro', 'macro', 'weighted']   

    The fp, tp, are computed following 


    """
    assert average in [None, 'micro', 'macro', 'weighted']

    # An array where for each class the specified quantities are computed
    df_confmat = online_confusion_matrix(y_true, y_pred, times, n_classes)
    from pyadlml.ml_viz import plotly_online_confusion_matrix
    plotly_online_confusion_matrix(conf_mat=df_confmat, scale='log').write_html('test.html')

    if average == 'micro' or average is None:
        confmat = df_confmat.values

        """
        Normally:
            A FP for class i counts when i is not (j), but i was predicted (j!=i)
            A FN for class i counts when i is but j was predicted (j!=i)
            If there is a FP for class i it counts as FN for class j and vice versa,
            therefore FP = FN in multiclass settings
        Since each sample has a weight, a specific duration a FN 
        """
        return np.diag(confmat).sum() / confmat.sum()

    elif average == 'macro':
        # Extract true as columns and pred as rows in nanosecond resolution
        # c_ij = true class i when predicted class is j
        confmat = df_confmat.values
        tp = np.diag(confmat)

        # Sum over is sth. else but was mistaken as c
        # fp = confmat.sum(1) - tp

        # Sum over is c but was predicted as sth. else
        # recovers y_true with df.groupby('y_true')['diff'].sum()
        fn = confmat.sum(0) - tp

        # All predictions that do not include c
        # tn = confmat.sum() - (fp + fn + tp)

        # Array of per class ppvs?
        score = _save_divide(tp, (tp + fn))

        # Normalize by each class by its
        normalized_score = (score*(1/n_classes)).sum(-1)
        return normalized_score
    else:
        raise NotImplementedError

def _slice_categorical_stream(df1, df2, first_ts=None, last_ts=None):
    """ Takes two categorical streams and chops both up with respect to
        each other timestamps 

    Parameters
    ----------
    df1 : pd.Dataframe
        [time, c1]
    df2 : pd.Dataframe
        [time, c1]

    Returns
    -------
    [time, c1, c2]
    """
    lbl1_col = set(df1.columns).difference([TIME]).pop()
    lbl2_col = set(df2.columns).difference([TIME]).pop()
    df1 = df1.rename(columns={lbl1_col: ACTIVITY})
    df2 = df2.rename(columns={lbl2_col: ACTIVITY})

    if first_ts is None:
        first_ts = max(df2.loc[0, TIME], df1.loc[0, TIME])
        first_ts -= pd.Timedelta('1ms')
    if last_ts is None:
        last_ts = min(df2.loc[df2.index[-1], TIME],
                      df1.loc[df1.index[-1], TIME])
        last_ts += pd.Timedelta('1ms')

    df1_tmp = df1.copy()
    df2_tmp = df2.copy()
    df1_tmp[ACTIVITY] = np.nan
    df2_tmp[ACTIVITY] = np.nan

    df2 = pd.concat([df2, df1_tmp], ignore_index=True, axis=0)\
            .sort_values(by=TIME)\
            .reset_index(drop=True)

    df1 = pd.concat([df1, df2_tmp], ignore_index=True, axis=0)\
            .sort_values(by=TIME)\
            .reset_index(drop=True)

    df1[ACTIVITY] = df1[ACTIVITY].ffill()
    df2[ACTIVITY] = df2[ACTIVITY].ffill()

    df1 = df1[(first_ts < df1[TIME]) & (df1[TIME] < last_ts)]
    df2 = df2[(first_ts < df2[TIME]) & (df2[TIME] < last_ts)]

    df = df1.copy()
    df[lbl2_col] = df2[ACTIVITY]
    df = df.rename(columns={ACTIVITY: lbl1_col})

    return df.copy()


def online_confusion_matrix(y_true: pd.DataFrame, y_pred: np.ndarray, times: np.ndarray,
                            n_classes: int, average=''):
    """
    Rows are predictions and columns are true values
    c_ij = predicted class is i at i-th row and true class j

    Parameters
    ----------
    y_true : pd.DataFrame


    Returns
    -------
    cm : pd.DataFrame

    """
    if is_activity_df(y_true):

        y_pred, times = y_pred.squeeze(), times.squeeze()

        # From [st, et, a1] -> [t, a1]
        df = y_true.copy()
        df = df.rename(columns={START_TIME: TIME})

        # Fill up with other
        df_other = df[[END_TIME, ACTIVITY]].copy()
        df_other.loc[:, ACTIVITY] = OTHER
        df_other = df_other.rename(columns={END_TIME: TIME})
        df = df.drop(columns=END_TIME)
        df = pd.concat([df, df_other.iloc[:-1]], ignore_index=True, axis=0) \
               .sort_values(by=TIME) \
               .reset_index(drop=True)
        df['diff'] = df[TIME].shift(-1) - df[TIME]
        mask_invalid_others = (df['diff'] < OTHER_MIN_DIFF) & (
            df[ACTIVITY] == OTHER)
        df = df[~mask_invalid_others][[TIME, ACTIVITY]]

        # Add the ending
        df = pd.concat([df, pd.Series({
            TIME: y_true.at[y_true.index[-1], END_TIME],
            ACTIVITY: y_true.at[y_true.index[-1], ACTIVITY]}
        ).to_frame().T])

        df_y_pred = pd.DataFrame({TIME: times, 'y_pred': y_pred})
        df_y_true = df.copy().rename(columns={ACTIVITY: 'y_true'})

        df_y_true = df_y_true.sort_values(by=TIME).reset_index(drop=True)
        df_y_pred = df_y_pred.sort_values(by=TIME).reset_index(drop=True)

        # if prediction is larger than ground truth activities pad gt with 'other'
        if df_y_pred[TIME].iat[0] < df_y_true[TIME].iat[0]:
            df_y_true = pd.concat([df_y_true,
                                   pd.Series(
                                       {TIME: df_y_pred.at[0, TIME], 'y_true':'other'})
                                   .to_frame().T], axis=0, ignore_index=True)

        if df_y_true[TIME].iat[-1] < df_y_pred[TIME].iat[-1]:
            df_y_true = pd.concat([df_y_true,
                                   pd.Series(
                                       {TIME: df_y_pred.at[df_y_pred.index[-1], TIME], 'y_true':'other'})
                                   .to_frame().T], axis=0, ignore_index=True)

        # if prediction frame is smaller than ground truth clip gt to prediction
        if df_y_pred[TIME].iat[-1] < df_y_true[TIME].iat[-1]:
            df_y_true[TIME].iat[-1] = df_y_pred[TIME].iat[-1]

        if df_y_pred[TIME].iat[0] < df_y_true[TIME].iat[0]:
            df_y_true[TIME].iat[0] = df_y_pred[TIME].iat[0]

        df = _slice_categorical_stream(df_y_pred,  df_y_true)

    else:
        y_true, y_pred, times = y_true.squeeze(), y_pred.squeeze(), times.squeeze()

        df = pd.DataFrame(data=[times, y_true, y_pred],
                          index=[TIME, 'y_true', 'y_pred']).T
        df[TIME] = pd.to_datetime(df[TIME])

    df['diff'] = df[TIME].shift(-1) - df[TIME]

    # TODO discuss
    # Impute last unknown time difference with mean
    # last_activity = df.at[df.shape[0]-1, 'y_true']
    # last_act_mean = df.groupby('y_true').mean().at[last_activity, 'diff']
    # df.at[df.shape[0], 'diff'] = last_act_mean
    df = df.iloc[:-1]

    cm = pd.crosstab(
        index=df['y_true'], columns=df['y_pred'], values=df['diff'], aggfunc=np.sum)

    cm = cm.replace({pd.NaT: pd.Timedelta('0s')})
    # TODO refactor, find out why cross tab creates 0 integers since the sum over values should be 
    cm = cm.replace({0: pd.Timedelta('0s')})

    missing_row = set(cm.columns) - set(cm.index)
    missing_col = set(cm.index) - set(cm.columns)
    if missing_row:
        # y_pred did not contain classes present in y_true
        for m_row in missing_row:
            new_row = pd.Series(name=m_row,
                                data=[pd.Timedelta('0s')]*len(cm.columns),
                                index=cm.columns
                                )
            cm = pd.concat([cm, new_row.to_frame().T], axis=0)

    if missing_col:
        # y_true did not contain classes present in y_pred
        for m_col in missing_col:
            cm[m_col] = pd.Series(name=m_col,
                                  data=[pd.Timedelta('0s')]*len(cm.index),
                                  index=cm.index
                                  )

        # Rearange columns and for diagonal to again matchup
        # when col or row was inserted above
        cm = cm.sort_index(ascending=True)
        cm = cm[cm.index.values]

    return cm



import pandas as pd
from pyadlml.constants import TIME, DEVICE, VALUE

def create_changepoint(df_devs):
    """

    Parameters
    ----------
    df_devs : pd.DataFrame
        A device dataframe, refer to guide

    Returns
    -------
    df : pd.DataFrame
    """

    # create binary vectors and set all changepoints to true
    df = df_devs.copy()
    df[VALUE] = True
    df = df.pivot(index=TIME, columns=DEVICE, values=VALUE)\
        .fillna(False)\
        .astype(int)\
        .reset_index()
    return df


def resample_changepoint(cp: pd.DataFrame, dt:str, use_dask=False) -> pd.DataFrame:
    """
    Resamples the changepoint representation with a given resolution

    Parameters
    ----------
    cp : pd.DataFrame
        A device dataframe in changepoint representation [TIME, dev1, dev2, ..., devn]

    dt : str
        

    Returns
    -------
    pd.DataFrame
        Resampled dataframe in changepoint representation
    """
    df = cp.copy() 
    df = df.sort_values(by=TIME).set_index(TIME)
    df = df.resample('1s', kind='timestamp').count()
    df[df > 1] = 1
    return df.reset_index()



from datetime import datetime, timezone
from typing_extensions import override
from pyadlml.dataset._core.acts_and_devs import label_data
from pyadlml.dataset.plot.matplotlib.util import save_fig
from pyadlml.pipeline import XOrYTransformer, XAndYTransformer, YTransformer
import numpy.ma as ma
from pyadlml.constants import ACTIVITY, DEVICE, END_TIME, OTHER, START_TIME, TIME, VALUE
from sklearn.base import BaseEstimator, TransformerMixin
from scipy.stats.kde import gaussian_kde
import dask
import pandas as pd
import numpy as np
from plotly.subplots import make_subplots
from pyadlml.dataset.plot.util import  \
    fmt_seconds2time, \
    fmt_seconds2time_log

class Windows(BaseEstimator, TransformerMixin):
    REP_M2M = 'many-to-many'
    REP_M2O = 'many-to-one'

    def __init__(self, rep, window_size, stride):
        self.rep = rep
        self.window_size = window_size
        self.stride = stride


    
    @override
    def nr_activities_per_win(self, y):
        raise NotImplementedError

    @save_fig    
    def plot_nr_activities_per_win(self, y, scale='linear', file_path=None):
        """ How many different activities are present in a window.
        
        Parameters
        ----------
        y: pd.Series or np.ndarray
            Including the activities and the timestamps ['time', 'activities']

        """
        import matplotlib.pyplot as plt
        win_size_str = int(self.window_size) if isinstance(self.window_size, float) else self.window_size
        title = f'Activities per window (size={win_size_str})'

        df = self.nr_activities_per_win(y)
        x, y = df['activity per window'], df['count']
        fig, ax = plt.subplots(1,1)
        ax.set_title(title)
        ax.bar(x, y)
        ax.xaxis.set_ticks(x)

        if scale == 'log':
            ax.set_yscale('log')
            ax.set_ylim(0, 10**np.ceil(np.log10(y.max())))

        return fig

    @classmethod
    def _gen_default_window_sizes(cls):
        return np.arange(4, 2000, 100) 

    @classmethod
    def plot_winsize_vs_activities(cls, df_devs, df_acts, window_sizes=None, scale='linear', n_bins=100):
        """


        """
        other = True
        window_sizes = cls._gen_default_window_sizes() if window_sizes is None else window_sizes

        title = 'Window size vs. activities'
        time_unit = 's'

        x_labeled = label_data(df_devs, df_acts, other=other)[[TIME, ACTIVITY]]
        if not other:
            x_labeled = x_labeled.dropna(axis=0)\
                                .reset_index(drop=True)

        max_acts, _ = cls(window_sizes=max(window_sizes)).activities_hist(x_labeled)

        z = np.zeros((max_acts.max(), len(window_sizes)), dtype=np.float32)
        for j, s in enumerate(window_sizes):
            x, y = cls(window_sizes=s).activities_hist(x_labeled)
            z[x-1, j] = y


        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(1,1)
        pcm = ax.pcolormesh(z, cmap='viridis', label='win_lengths', norm=scale)
        x_tick_pos = np.arange(0, len(window_sizes), np.ceil(len(window_sizes)*0.1), dtype=np.int32)
        ax.set_xticks(x_tick_pos, labels=window_sizes[x_tick_pos])

        from pyadlml.dataset.plot.matplotlib.util import make_axes_locatable
        ax.set_xlabel('#events per window')
        ax.set_ylabel('#activities per window')
        ax.set_title(title)


        divider = make_axes_locatable(ax)
        cax = divider.new_horizontal(size="4%", pad=0.1, pack_start=False)
        fig.add_axes(cax)
        fig.colorbar(pcm, cax=cax, orientation="vertical")
        return fig



class ExplicitWindow(BaseEstimator, TransformerMixin, XOrYTransformer):
    """
    https://www.mdpi.com/1424-8220/21/18/6037

    Divide the data stream into segments by 

    """
    def __init__(self, rep: str = 'many-to-many'):
        TransformerMixin.__init__(self)
        XOrYTransformer.__init__(self)
        self.rep = rep

    def fit(self, X, y):
        assert self.rep in ['many-to-many', 'many-to-one']
        return self

    def fit_transform(self, X, y):
        self.fit(X,y)
        return self.transform(X,y)

    @XOrYTransformer.x_or_y_transform
    def transform(self, X: pd.DataFrame, y:pd.DataFrame) -> pd.DataFrame:
        """

        Parameters
        ----------
        X: pd.DataFrame

        y: pd.DataFrame


        .. Note 
        -------

        Sequences are padded with NaNs to the length of the longest sequence in order to form a tensor.


        Returns
        -------
        X: np.ndarray with shape (S, F, T)
            where S is the #sequences, F is the #features, and T is the maxmimum sequence length
        y: np.ndarray with shape (S, T)
            where S is the #sequences and T is the maximum sequence length

        """
        assert len(y.columns) == 1

        y = y.copy().reset_index(drop=True)
        X = X.copy().reset_index(drop=True)

        # Assign each sequence a unique number
        y['tmp'] = (y[ACTIVITY] != y[ACTIVITY].shift(1)).cumsum()
        S = y['tmp'].iat[-1]
        T = y['tmp'].value_counts().max()
        F = len(X.columns)

        Xt = np.full((S, F, T), np.nan, dtype='object')
        if self.rep == 'many-to-many':
            yt = np.full((S, T), np.nan, dtype='object')
        else:
            yt = np.full((S), '', dtype='object')

        for s, (_, grp) in enumerate(y.groupby('tmp')):
            X_tmp = np.swapaxes(X.loc[grp.index, :].values, 0,1)
            Xt[s, :, :X_tmp.shape[-1]] = X_tmp
            if self.rep == 'many-to-many':
                y_tmp = y.loc[grp.index, y.columns[:-1]].values.squeeze(-1)
                yt[s, :y_tmp.shape[-1]] = y_tmp
            else:
                yt[s] = y.at[grp.index[0], ACTIVITY]

        return Xt, yt

class TimeWindow(Windows):
    """ Divide data stream into time segments with a regular interval.

    Note
    ----
    Also works with regular sampled data (dt='xs').
    Van Kasteren et al. recommend 60s. 

    """
    def __init__(self, window_size : str, stride: str = None, rep: str ='many-to-many',  drop_empty_intervals=True):
        TransformerMixin.__init__(self)
        XOrYTransformer.__init__(self)
        Windows.__init__(self, rep, window_size, stride)
        self.drop_empty_intervals = drop_empty_intervals

    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X, y)

    def fit(self, X=None, y=None):

        # Bring params into right format
        assert self.rep in [self.REP_M2M, self.REP_M2O]
        self.window_size = pd.Timedelta(self.window_size)
        self.stride = pd.Timedelta(self.stride) if self.stride is not None else self.window_size
        return self

    @XOrYTransformer.x_or_y_transform
    def transform(self, X, y=None) -> np.ndarray:
        """
        
        Parameters 
        ----------- (S, T, F) and (S, T)
        """
        # TODO refactor add conversion for different input types
        assert isinstance(X, pd.DataFrame) or X is None
        assert isinstance(y, pd.DataFrame) or y is None

        df = X.copy().sort_values(by=TIME)
        st = X[TIME].iloc[0] - pd.Timedelta('1s')
        et = X[TIME].iloc[-1] + pd.Timedelta('1s')
        win_st = st

        X_list = []
        y_list = []
        max_seq_length = 0
        i = 0


        st_windows = pd.date_range(st, et-self.window_size, freq=self.stride)
        if self.drop_empty_intervals:
            times = df[TIME].copy()
            while win_st + self.window_size <= et:

                win = (win_st, win_st + self.window_size)
                # Important the interval is [st,et) right closed
                event_idxs = df[(win[0] <= df[TIME]) & (df[TIME] < win[1])].index
                
                if not event_idxs.empty:
                    X_list.append(X.iloc[event_idxs].copy())
                    if y is not None:
                        y_list.append(y.iloc[event_idxs].copy())
                    max_seq_length = max(max_seq_length, len(X_list[i]))
                    win_st = win_st + self.stride
                    i += 1
                else:
                    # Get the next first window that covers the next event 
                    next_event_time = times[win_st < times].iloc[0]
                    win_min_idx_not_containing_ev = (st_windows <= next_event_time - self.window_size).cumsum().max() - 1
                    if win_min_idx_not_containing_ev == len(st_windows) - 1:
                        # When there is no event for the enxt windows 
                        break
                    win_st = st_windows[win_min_idx_not_containing_ev + 1]
        else: 
            for win_st in st_windows:
                win = (win_st, win_st + self.window_size)
                event_idxs = df[(win[0] <= df[TIME]) & (df[TIME] < win[1])].index
                
                X_list.append(X.iloc[event_idxs].copy())
                if y is not None:
                    y_list.append(y.iloc[event_idxs].copy())
                max_seq_length = max(max_seq_length, len(X_list[i]))
                i += 1


        S = len(X_list)
        F = len(X.columns)
        T = max_seq_length
        Xt = np.full((S, T, F), np.nan, dtype='object')

        if y is not None:
            Fy = 1 if len(y.shape) == 1 else y.shape[-1]
            if self.rep == 'many-to-many':
                yt = np.full((S, T, Fy), np.nan, dtype='object')
            else:
                yt = np.full((S, Fy), np.nan, dtype='object')

        for s in range(len(X_list)):
            X_tmp = X_list[s].values
            Xt[s, :X_tmp.shape[0], :] = X_tmp

            if y is not None:
                y_tmp = y_list[s].values
                if y_tmp.size > 0:
                    if self.rep == 'many-to-many':
                        y_tmp = y_tmp.squeeze(-1)
                        yt[s, :y_tmp.shape[0], :] = y_tmp
                    else:
                        yt[s, :] = y_tmp[-1]

        yt = None if y is None else yt
        return Xt, yt 

    @classmethod
    def nr_windows(cls, X, win_size, stride='10s', drop_empty_intervals=False):
        df = X.copy().sort_values(by=TIME)
        st = X[TIME].iloc[0] - pd.Timedelta('1s')
        et = X[TIME].iloc[-1] + pd.Timedelta('1s')
        st_windows = pd.date_range(st, et-win_size, freq=stride)
        return st_windows

    def nr_activities_per_win(self, y):
        """ How many different activities are present in a window

        Parameters
        ----------
        y: pd.DataFrame
            A table with at least one column named 'activity' and 'time'

        Returns
        -------
        pd.DataFrame
        """
        assert (y.columns == [TIME, ACTIVITY]).all()
        self.rep = self.REP_M2M
        contains_nans = y[ACTIVITY].isna().any()
        #if contains_nans:
        #    act2int = { v:k for k, v in enumerate(y[ACTIVITY].unique())} 
        #    from pandas._libs.tslibs.nattype import NaT
        #    #act2int[NaT] = np.inf
        #    y = self.transform(X=y)
        #    y = np.vectorize(act2int.get)(y).astype(np.int32)
        #    y = np.where(y == act2int[NaT], np.inf, y)
        #    y = ma.masked_invalid(y)
        #    counts_per_row = (y[:,1:] != y[:,:-1]).sum(axis=1)+1
        #    x, y = np.unique(counts_per_row, return_counts=True)
        #    df = pd.DataFrame(columns=['activity per window', 'count'], data=np.stack([x,y]).T)
        #    return df.sort_values(by='activity per window')
        #    
        #else:
        if True:
            act2int = { v:k for k, v in enumerate(y[ACTIVITY].unique())} 
            yt = self.transform(X=y)    # S, T, 2 (timestamps, activities) in feature_y dim
            yt = yt[:, :, 1] # remove timestamps -> (S, T)

            # Hotfix nan issue
            yt = np.where(pd.isnull(yt), -1, yt)
            act2int[-1] = -1
            yt = np.vectorize(act2int.get)(yt).astype(np.int32)
            yt = np.where(yt == -1, np.inf, yt)
            yt = ma.masked_invalid(yt)
            counts_per_row = (yt[:,1:] != yt[:,:-1]).sum(axis=1)+1
            bins, counts = np.unique(counts_per_row, return_counts=True)
            return pd.DataFrame(columns=['activity per window', 'count'], data=np.stack([bins, counts]).T)


    @save_fig
    def plot_nr_events_per_window(self, times, y_scale='linear', n_bins=20, file_path=None):
        """ Plot the number of events per window

        Parameters
        ----------
        times : pd.DataFrame
            TODO
        y_scale : str, one of ['linear', 'log'], default='linear'
            TODO
        n_bins : int, default=20
            TODO
        file_path : None or str

        Returns
        -------
        fig or None
        """

        # Get nr events per window
        yt = self.transform(X=times) 
        yt = yt[:, :, 1] # Remove timestamps -> (S, T)
        counts = np.count_nonzero(~pd.isnull(yt), axis=1)

        title = f'#Events per window (size={self.window_size})'

        import matplotlib.pyplot as plt
        from pyadlml.dataset.plot.matplotlib.util import plot_hist
        fig, ax = plt.subplots(1,1)

        bins = np.linspace(counts.min(), counts.max(), n_bins+1)

        hist, _ = np.histogram(counts, bins=bins)
        mean = counts.mean() 
        median = np.median(counts)

        if y_scale == 'log':
            y_max = ax.set_ylim(1, 10**np.ceil(np.log10(hist.max())))
        else:
            y_max = hist.max()
            y_max += y_max*0.05

        ax.hist(counts, bins, label='')
        ax.set_title(title, y=1.08)
        ax.set_xlabel('#events')
        ax.set_ylabel('count')

        if y_scale == 'log':
            ax.set_yscale('log')
            ax.set_ylim(1, top=y_max)
        else:
            ax.set_ylim(0, y_max)

        ax.plot([mean]*2, [0, y_max], label=f'mean {mean:.2f}')
        ax.plot([median]*2, [0, y_max], label=f'median {median:.2f}')

        ax.legend()

        return fig



    def activities_hist(self, Xt):
        """
        """
        # Retrieve windows and only activities
        #Xt = Xt[:, None, :] if win_size == 1 else Xt

        Xt = Xt[:,:,1]

        # Get histogram
        counts_per_row = (Xt[:,1:] != Xt[:,:-1]).sum(axis=1)+1
        x, y = np.unique(counts_per_row, return_counts=True)
        return x, y

    @classmethod
    def _gen_default_window_sizes(cls):
        return np.arange(4, 2000, 100) 

    @classmethod
    @save_fig
    def plot_winsize_vs_activities(cls, df_devs, df_acts, window_sizes=None, scale='linear', other=False, stride='10s', use_dask=False, file_path=None):
        """
        
        Parameters
        ----------

        """
        window_sizes = pd.timedelta_range('1h', '20D', freq='1h') if window_sizes is None else window_sizes

        title = 'Window size vs. activities'
        scale = 'log'

        x_labeled = label_data(df_devs, df_acts, other=other)[[TIME, ACTIVITY]]

        if not other:
            x_labeled = x_labeled.dropna(axis=0)\
                                .reset_index(drop=True)

        max_acts = cls(window_size=max(window_sizes), stride=stride).fit()\
                .nr_activities_per_win(x_labeled)['activity per window']\
                .max()
        if use_dask:
            z = np.zeros((max_acts.max(), len(window_sizes)), dtype=np.float32)
            res = []
            def fun(s, x):
                df = cls(window_size=s, stride=stride).fit().nr_activities_per_win(x)
                return df['count'], df['activity per window']
            x_labeled = dask.delayed(x_labeled)
            for s in window_sizes:
                res.append(dask.delayed(fun)(s, x_labeled))

            res = dask.compute(*res)
            for xi, (zi, y) in enumerate(res):
                z[y-1, xi] = zi

        else:
            z = np.zeros((max_acts.max(), len(window_sizes)), dtype=np.float32)
            for xi, s in enumerate(window_sizes):
                df = cls(window_size=s, stride=stride).fit().nr_activities_per_win(x_labeled)
                zi, y = df['count'], df['activity per window']
                z[y-1, xi] = zi


        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(1,1)
        pcm = ax.pcolormesh(z, cmap='viridis', label='win_lengths', norm=scale)
        x_tick_pos = np.arange(0, len(window_sizes), np.ceil(len(window_sizes)*0.1), dtype=np.int32)
        ax.set_xticks(x_tick_pos + 1 , labels=window_sizes[x_tick_pos])

        from pyadlml.dataset.plot.matplotlib.util import make_axes_locatable
        ax.set_xlabel('window $dt$')
        ax.set_ylabel('#activities per window')
        ax.set_title(title)


        divider = make_axes_locatable(ax)
        cax = divider.new_horizontal(size="4%", pad=0.1, pack_start=False)
        fig.add_axes(cax)
        fig.colorbar(pcm, cax=cax, orientation="vertical")
        ax.tick_params(axis='x', rotation=-45)
        plt.tight_layout()
        return fig

    def nr_events_per_win(self, times):
        """

        # Get nr events per window
        # Pick first value after timestamps. The value is irrelevant just that it is not None
        # -> (S, T, F)
        """
        yt = self.transform(X=times) 
        yt = yt[:, :, 1] # Remove timestamps -> (S, T)
        counts_per_row = np.count_nonzero(~pd.isnull(yt), axis=1)
        bins, counts = np.unique(counts_per_row, return_counts=True)
        return pd.DataFrame(columns=['events per window', 'count'], data=np.stack([bins, counts]).T)

    def get_event_rate(self, X):
        """
        
        """
        self.drop_empty_intervals = False
        st = X[TIME].iloc[0] - pd.Timedelta('1s')
        et = X[TIME].iloc[-1] + pd.Timedelta('1s')
        st_windows = pd.date_range(st, et-self.window_size, freq=self.stride)
        st_windows = st_windows + self.window_size
        # -> (S, T, F)
        Xt = self.transform(X=X) 
        counts_per_row = np.count_nonzero(~pd.isnull(Xt[:,:,0]), axis=1)
        df = pd.DataFrame([counts_per_row, st_windows]).T
        df.columns = [VALUE, TIME]
        df[DEVICE] = 'Time window: Event rate'

        return df[[TIME, DEVICE, VALUE]].sort_values(by=TIME).reset_index(drop=True)



    @classmethod
    @save_fig
    def plot_winsize_vs_events(cls, df_devs, window_sizes=None, scale='log', stride='10s', use_dask=False, file_path=None):
        """
        
        Parameters
        ----------

        """
        window_sizes = pd.timedelta_range('1h', '20D', freq='1h') if window_sizes is None else window_sizes

        title = 'Window size vs. events'

        x_labeled = df_devs

        max_acts = int(cls(window_size=max(window_sizes), stride=stride).fit()\
                .nr_events_per_win(df_devs)['events per window']\
                .max())
        if use_dask:
            z = np.zeros((max_acts, len(window_sizes)), dtype=np.float32)
            res = []
            def fun(s, x):
                df = cls(window_size=s, stride=stride).fit().nr_events_per_win(x)
                return df['count'], df['events per window']
            x_labeled = dask.delayed(x_labeled)
            for s in window_sizes:
                res.append(dask.delayed(fun)(s, x_labeled))

            res = dask.compute(*res)
            for xi, (zi, y) in enumerate(res):
                z[y-1, xi] = zi

        else:
            z = np.zeros((max_acts.max(), len(window_sizes)), dtype=np.float32)
            for xi, s in enumerate(window_sizes):
                df = cls(window_size=s, stride=stride).fit().nr_activities_per_win(x_labeled)
                zi, y = df['count'], df['events per window']
                z[y-1, xi] = zi


        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(1,1)
        pcm = ax.pcolormesh(z, cmap='viridis', label='win_lengths', norm=scale)
        x_tick_pos = np.arange(0, len(window_sizes), np.ceil(len(window_sizes)*0.1), dtype=np.int32)
        ax.set_xticks(x_tick_pos + 1 , labels=window_sizes[x_tick_pos])

        from pyadlml.dataset.plot.matplotlib.util import make_axes_locatable
        ax.set_xlabel('window $dt$')
        ax.set_ylabel('#events per window')
        ax.set_title(title)

        divider = make_axes_locatable(ax)
        cax = divider.new_horizontal(size="4%", pad=0.1, pack_start=False)
        fig.add_axes(cax)
        fig.colorbar(pcm, cax=cax, orientation="vertical")
        ax.tick_params(axis='x', rotation=-45)
        plt.tight_layout()
        return fig

class FuzzyTimeWindows():
    """
    """
    def __init__(self):
        raise NotImplementedError


class EventWindow(Windows, XOrYTransformer):
    """ Generate subsequences from a 




    Many-To-Many
    ^^^^^^^^^^^^

    To get *many-to-many* batches use the window size :math:`w` to split the data

    .. math::
        f(X_{N,K},y_{N}) \rightarrow (X_{W, N,K}, y_{W, N})

    In addition you can specify a stride.


    .. code:: python

        from pyadlml.dataset import fetch_kasteren
        from pyadlml.preprocessing import EventWindows
        data = fetch_kasteren()

        raw = StateVectorEncoder(encode='raw', t_res='10s')\
              .fit_transform(data.df_devices)
        labels = LabelEncoder().fit_transform(raw, data.df_activities)

        X = raw.values
        y = labels.values

        X, y = EventWindows(rep='many-to-many', window_length=10, stride=2)\
               .fit_transform(X, y)

    Many-To-One
    ^^^^^^^^^^^

    .. math::
        f(X_{N,K},y_{N}) \rightarrow (X_{W, N, K}, y_{N})


    .. code:: python

        from pyadlml.preprocessing import SequenceDicer

        raw = StateVectorEncoder(encode='raw', t_res='10s')\
              .fit_transform(data.df_devices)
        labels = LabelEncoder().fit_transform(raw, data.df_activities)

        X = raw.values
        y = labels.values

        X, y = SequenceDicer(rep='many-to-one', window_length='20s')\
               .fit_transform(X, y)

    """
    def __init__(self, rep: str ='many-to-many', window_size: int =10, stride: int=1, use_stride_trick=False, dtype=None):
        """
        Parameters
        ----------
        rep: str 
        window_size: int
        stride: int

        """

        TransformerMixin.__init__(self)
        XOrYTransformer.__init__(self)
        Windows.__init__(self, rep, window_size, stride)
        self.use_stride_trick = use_stride_trick
        self.dtype = dtype
    
    def fit(self, X, y=None):
        self.feature_names_in_ = X.columns if isinstance(X, pd.DataFrame) else None
        self.X_shape_ =(self._calc_new_N(X.shape[0]), self.window_size, *X.shape[1:])
        return self


    def fit_transform(self, X, y=None):
        """
        Parameters
        ----------
        X : np.array
            Some kind of numpy array or pandas Dataframe
        y : np.array

        Returns
        -------
        x : np.array of shape ()
        y : np.array
            todo
        """
        self.fit(X, y)
        return self.transform(X, y)

    @XOrYTransformer.x_or_y_transform
    def transform(self, X=None, y=None):
        """

        """
        assert self.rep in [self.REP_M2M, self.REP_M2O]

        if X is not None:
            assert self.window_size < len(X), f'Length X: {len(X)} >= {self.window_size} window size!'
            assert self.stride < len(X)

            X = self._transform_X(X)

        if y is not None:
            assert self.window_size < len(y)
            assert self.stride < len(y)

            y = self._transform_Y(y)

        return X, y

    def _calc_new_N(self, n_old):
        return int(np.floor((n_old-self.window_size)/self.stride)+1)

    def _transform_X(self, X):
        """

        """
        if isinstance(X, pd.DataFrame) or isinstance(X, pd.Series):
            X = X.to_numpy()

        n_prime = self._calc_new_N(X.shape[0])
        new_shape =(n_prime, self.window_size, *X.shape[1:])

        dtype = X.dtype if self.dtype is None else self.dtype
        from numpy.core._exceptions import _ArrayMemoryError
        if self.use_stride_trick:
            new_shape = (X.shape[0] - self.window_size + 1, self.window_size, *X.shape[1:])
            strides = (X.strides[0],) + X.strides 
            res = np.lib.stride_tricks.as_strided(X, shape=new_shape, strides=strides, 
                                                  writeable=False)
            return res
        else:
            try:
                res = np.zeros(shape=new_shape, dtype=dtype)
            except _ArrayMemoryError:
                from pathlib import Path
                self.fp_memmap_ = Path("/tmp/mmemap.dat")
                print('Warning! memcache boolean dtype used. This may be wrong')
                res = np.memmap(self.fp_memmap_, mode='w+', shape=new_shape, dtype=np.bool_)

            for r, i in enumerate(range(0, n_prime*self.stride, self.stride)):
                res[r] = X[i:i+self.window_size]
                #res[r, :, :] = X[i:i+self.window_size, :]

            return res.squeeze()

    def _transform_Y(self, y):
        """


        """

        if isinstance(y, pd.DataFrame):
            y = y.to_numpy().squeeze()
        elif isinstance(y, pd.Series):
            y = y.to_numpy()

        dtype = y.dtype if self.dtype is None else self.dtype
        if self.rep == self.REP_M2M:
            n_prime = self._calc_new_N(y.shape[0])
            new_shape = [n_prime, self.window_size, *y.shape[1:]]

            res = np.zeros(shape=new_shape, dtype=dtype)

            for r, i in enumerate(range(0, n_prime*self.stride, self.stride)):
                res[r, :] = y[i: i+self.window_size]

            return res.squeeze()

        elif self.rep == self.REP_M2O:
            res = y[np.arange(self.window_size-1, y.shape[0], self.stride)]
            return res.squeeze().astype(dtype)

    def nr_activities_per_win(self, y):
        """ How many different activities are present in a window

        Parameters
        ----------
        y: pd.DataFrame
            A table with at least one column named 'activity' and 'time'

        Returns
        -------
        pd.DataFrame
        """
        assert (y.columns == [TIME, ACTIVITY]).all()
        self.rep = self.REP_M2M
        contains_nans = y[ACTIVITY].isna().any()
        if contains_nans:
            act2int = { v:k for k, v in enumerate(y[ACTIVITY].unique())} 
            from pandas._libs.tslibs.nattype import NaT
            #act2int[NaT] = np.inf
            y = self.transform(X=y)[:,:,1]
            y = np.vectorize(act2int.get)(y).astype(np.int32)
            y = np.where(y == act2int[NaT], np.inf, y)
            y = ma.masked_invalid(y)

            counts_per_row = (y[:,1:] != y[:,:-1]).sum(axis=1)+1
            x, y = np.unique(counts_per_row, return_counts=True)
            df = pd.DataFrame(columns=['activity per window', 'count'], data=np.stack([x,y]).T)
            return df.sort_values(by='activity per window')
            
        else:
            act2int = { v:k for k, v in enumerate(y[ACTIVITY].unique())} 
            y = self.transform(X=y[ACTIVITY])
            y = np.vectorize(act2int.get)(y)
            counts_per_row = (y[:,1:] != y[:,:-1]).sum(axis=1)+1
            x, y = np.unique(counts_per_row, return_counts=True)
            return pd.DataFrame(columns=['activity per window', 'count'], data=np.stack([x,y]).T)
            


    def get_coverage(self, times, n_samples=1000000, dt='5s'):
        """ Function of amount of intervals covering at times. A increment and decrement
            process based on interval start end end times. 

        Parameters
        ----------
        times : pd.Series or np.ndarray

        n_samples : int
            The number of monte carlo samples to draw

        Returns
        -------
        pd.DataFrame
            A device dataframe with columns ['time', 'device', 'value']
        """

        # Get intervals from windows
        x = self.fit_transform(times)
        df = pd.DataFrame(data=np.stack([x[:,0], x[:,-1]]).T, columns=[START_TIME, END_TIME])
        df = df.sort_values(by=START_TIME).reset_index(drop=True)
        steps = (df[END_TIME].iloc[-1] - df[START_TIME].iloc[0])//pd.Timedelta(dt)

        assert n_samples > len(df)
        n_samples_per_int = n_samples//len(df)


        # Transform to unix time 
        df = df.applymap(lambda x: x.replace(tzinfo=timezone.utc).timestamp())

        points = np.zeros((n_samples_per_int*len(df)), dtype=np.float32)
        # Sample point from intervals
        for i in range(df.shape[0]):
            intv = df.iloc[i]
            unx_sample = np.random.uniform(intv[START_TIME], intv[END_TIME], n_samples_per_int)
            j = i*n_samples_per_int
            points[j:j+n_samples_per_int] = unx_sample

        # Kernel density estimation 
        x = np.linspace(df[START_TIME].min(), df[END_TIME].max(), steps)
        y = gaussian_kde(points).pdf(x)

        df_res = pd.DataFrame(data=np.vectorize(datetime.utcfromtimestamp)(x), columns=[TIME])
        df_res[TIME] = pd.to_datetime(df_res[TIME])
        df_res[VALUE] = y
        df_res[DEVICE] = 'EventWin Coverage'
        return df_res


    def get_durations_per_window(self, times): 
        x = self.fit_transform(times)
        time_unit = 's'
        dur = (x[:,-1] - x[:,0])/np.timedelta64(1, time_unit)
        return dur

    def plot_duration_per_window(self, times, y_scale='linear', x_scale='linear', n_bins=20):
        dur = self.get_durations_per_window(times)

        title = f'Window lengths (size={int(self.window_size)})'

        import matplotlib.pyplot as plt
        from pyadlml.dataset.plot.matplotlib.util import plot_hist
        fig, ax = plt.subplots(1,1)
        
        
        if x_scale == 'log':
            ax.set_xscale('log')
            # Ensures bins are equally spaced
            bins = np.logspace(np.log10(dur.min()), np.log10(dur.max()), n_bins)
        else:
            bins = np.linspace(dur.min(), dur.max(), n_bins+1)

        hist, _ = np.histogram(dur, bins=bins)
        mean = dur.mean() 
        median = np.median(dur)

        if y_scale == 'log':
            y_max = ax.set_ylim(1, 10**np.ceil(np.log10(hist.max())))
        else:
            y_max = hist.max()
            y_max += y_max*0.05

        ax.hist(dur, bins, label='win_lengths')
        ax.set_title(title, y=1.08)
        ax.set_xlabel('time in seconds')
        ax.set_ylabel('count')

        if y_scale == 'log':
            ax.set_yscale('log')
            ax.set_ylim(1, top=y_max)
        else:
            ax.set_ylim(0, y_max)

        ax.plot([mean]*2, [0, y_max], label='mean')
        ax.plot([median]*2, [0, y_max], label='median')

        ax.legend()

        from pyadlml.dataset.plot.matplotlib.util import add_top_axis_time_format
        add_top_axis_time_format(ax)
        return fig


    # TODO refactor
    #def plotly_windows_into_acts_and_devs(self, df_devs, df_acts):
    #    """

    #    """

    #    # Get times 
    #    Xt = self.fit_transform(df_devs) # (S, T, 3)


    #    from pyadlml.plot import plotly_activities_and_devices
    #    fig = plotly_activities_and_devices(df_devs, df_acts)

    #    # Plot times as vertical bars into acts_and_devs plot 
    #    for i in range(Xt.shape[0]):
    #        st = Xt[i][0, 0]
    #        et = Xt[i][-1, 0]
    #        fig.add_vline(x=st, line_width=2, line_color="Red", line_dash="dash")
    #        fig.add_vline(x=et, line_width=2, line_color="Red")

    #    return fig


    @classmethod
    def plot_winsize_vs_activities(cls, df_devs, df_acts, window_sizes=None, scale='linear', n_bins=100):
        """


        """
        other = True
        window_sizes = np.arange(4, 2000, 100) if window_sizes is None else window_sizes

        title = 'Window size vs. activities'
        time_unit = 's'

        x_labeled = label_data(df_devs, df_acts, other=other)[[TIME, ACTIVITY]]
        if not other:
            x_labeled = x_labeled.dropna(axis=0)\
                                .reset_index(drop=True)
        def get_histogram(win_size, x_labeled):
            # Retrieve windows and only activities
            Xt = cls(window_size=win_size).fit_transform(x_labeled)
            Xt = Xt[:, None, :] if win_size == 1 else Xt

            Xt = Xt[:,:,1]

            # Get histogram
            counts_per_row = (Xt[:,1:] != Xt[:,:-1]).sum(axis=1)+1
            x, y = np.unique(counts_per_row, return_counts=True)
            return x, y

        max_acts, _ = get_histogram(max(window_sizes), x_labeled)
        z = np.zeros((max_acts.max(), len(window_sizes)), dtype=np.float32)
        for j, s in enumerate(window_sizes):
            x, y = get_histogram(s, x_labeled)
            z[x-1, j] = y


        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(1,1)
        pcm = ax.pcolormesh(z, cmap='viridis', label='win_lengths', norm=scale)
        x_tick_pos = np.arange(0, len(window_sizes), np.ceil(len(window_sizes)*0.1), dtype=np.int32)
        ax.set_xticks(x_tick_pos, labels=window_sizes[x_tick_pos])

        from pyadlml.dataset.plot.matplotlib.util import make_axes_locatable
        ax.set_xlabel('#events per window')
        ax.set_ylabel('#activities per window')
        ax.set_title(title)


        divider = make_axes_locatable(ax)
        cax = divider.new_horizontal(size="4%", pad=0.1, pack_start=False)
        fig.add_axes(cax)
        fig.colorbar(pcm, cax=cax, orientation="vertical")
        return fig

    @classmethod
    def plot_winsize_vs_lengths(cls, X, y, window_sizes=None, z_scale='linear', y_scale='linear'):
        assert window_sizes[0] > 1, 'Length is not defined for window containing one timestamp'

        # Reduce X to only time dimension
        # TODO, refactor 
        x_times = X[TIME].values

        title = 'Window size vs. duration'
        cbarlabel = 'length' if z_scale == 'linear' else 'log length'
        n_bins = 100
        time_unit = 's'
        _ = None


        z = np.zeros((n_bins, len(window_sizes)), dtype=np.float32)
        dur_list = []
        dur_min = float('inf') 
        dur_max = float('-inf')
        for s in window_sizes:
            Xt = cls(window_size=s).fit_transform(x_times)
            # Get normalized time lengths
            Xt_dur = (Xt[:,-1] - Xt[:,0])/np.timedelta64(1, time_unit)
            
            dur_min = min(dur_min, Xt_dur.min())
            dur_max = max(dur_max, Xt_dur.max())
            dur_list.append(Xt_dur)

        # Make equal bin size from max to min
        # If log, Bins are equally sized in log space 
        if y_scale == 'log':
            dur_min, dur_max = np.log10(dur_min), np.log10(dur_max)
            bins = np.logspace(dur_min, dur_max, n_bins+1)
        else:
            bins = np.linspace(dur_min, dur_max, n_bins+1)


        for pos, d in enumerate(dur_list):
            hist, edges = np.histogram(d, bins=bins)
            z[:, pos] = hist

        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(1,1)
        X, Y = np.meshgrid(window_sizes, bins[:-1])
        pcm = ax.pcolormesh(X, Y, z, cmap='viridis', label='win_lengths', norm=z_scale)
        #x_tick_pos = np.arange(0, len(window_sizes), np.ceil(len(window_sizes)*0.1), dtype=np.int32)
        #ax.set_xticks(x_tick_pos, labels=window_sizes[x_tick_pos])
        #y_tick_pos = np.arange(0, len(bins), np.ceil(len(bins)*0.1), dtype=np.int32)
        #ax.set_yticks(y_tick_pos, labels=np.vectorize(lambda x: f'{x:.2f}')(bins[y_tick_pos]))
        if y_scale == 'log':
            ax.set_yscale('log')
        from mpl_toolkits.axes_grid1.inset_locator import inset_axes

        from pyadlml.dataset.plot.matplotlib.util import add_axis_time_format, make_axes_locatable
        add_axis_time_format(ax, 'right', scale='linear')
        ax.set_xlabel('#events per window')
        ax.set_ylabel('duration seconds')
        ax.set_title(title)

        divider = make_axes_locatable(ax)
        cax = divider.new_horizontal(size="4%", pad=0.7, pack_start=False)
        fig.add_axes(cax)
        fig.colorbar(pcm, cax=cax, orientation="vertical")
        return fig
        

    @classmethod
    def plotly_winsize_vs_length(cls, X, y, window_sizes=None, z_scale='linear', y_scale='linear'):
        """
        """
        import plotly.graph_objects as go
        window_sizes = np.arange(2, 1000, 10) if window_sizes is None else window_sizes

        assert window_sizes[0] > 1, 'Length is not defined for window containing one timestamp'

        # Reduce X to only time dimension
        # TODO, refactor 
        x_times = X[TIME].values

        title = 'Window size vs. window_length'
        cbarlabel = 'length' if z_scale == 'linear' else 'log length'
        n_bins = 100
        time_unit = 's'
        _ = None


        z = np.zeros((n_bins, len(window_sizes)), dtype=np.float32)
        dur_list = []
        dur_min = float('inf') 
        dur_max = float('-inf')
        for s in window_sizes:
            Xt = cls(window_size=s).fit_transform(x_times)
            # Get normalized time lengths
            Xt_dur = (Xt[:,-1] - Xt[:,0])/np.timedelta64(1, time_unit)
            
            dur_min = min(dur_min, Xt_dur.min())
            dur_max = max(dur_max, Xt_dur.max())
            dur_list.append(Xt_dur)

        # Make equal bin size from max to min
        # If log, Bins are equally sized in log space 
        #bin_gen = np.logspace if y_scale == 'log' else np.linspace
        #if y_scale == 'log':
        #    dur_min, dur_max = np.log10(dur_min), np.log10(dur_max)
        bin_gen = np.linspace
        bins = bin_gen(dur_min, dur_max, n_bins+1)


        for pos, d in enumerate(dur_list):
            hist, edges = np.histogram(d, bins=bins)
            z[:, pos] = hist

        z = np.log(z) if z_scale == 'log' else z

        fig = go.Figure(data=go.Heatmap(
                            x=window_sizes,
                            z=z,
                            colorscale='Viridis',
        ))
        if z_scale == 'log':
            tmpltstr = '%{x} event-length<br>' \
                     + ' -> %{customdata[1]} windows durating %{customdata[0]}.<extra></extra>'
            fig.data[0].hovertemplate = tmpltstr
            y_fmt = bins[:-1].copy().astype(object)

            for i in range(len(y_fmt)):
                if y_scale == 'log':
                    y_fmt[i] = fmt_seconds2time_log(y_fmt[i])
                else:
                    y_fmt[i] = fmt_seconds2time(y_fmt[i])

            y_fmt = np.repeat(y_fmt, len(window_sizes)).reshape(len(y_fmt), len(window_sizes))

            tmp = (np.exp(z.copy()) if z_scale == 'log' else z).astype(np.int32) 

            cd = np.array([y_fmt, tmp])
            fig['data'][0]['customdata'] = np.moveaxis(cd, 0, -1)

        #fig.update_yaxes(type='log')
        fig.update_layout(yaxis=dict(tickmode='array', tickvals=np.arange(len(bins[:-1])), ticktext=edges[:-1]))

        from pyadlml.dataset.plot.plotly.activities import _set_compact_title
        _set_compact_title(fig, title=title)
        return fig


    def construct_target_times(self, times: np.ndarray) -> np.ndarray:
        times = self._transform_Y(times)
        
        if self.rep == self.REP_M2M:
            # TODO take only last y-obs per sequence??? 
            # Or average y for each position when plotting 
            raise NotImplementedError
        
        return times


    def construct_X_at_target(self, X: pd.DataFrame, dev_slice=slice(None, None, None)) -> np.ndarray:
        """ Reconstruct the datapoints that are at the points
            where the target is evaluated

        Parameters
        ----------
        X: pd.DataFrame, (Tt, F)
            Transformed data right before windowing 

        Returns
        -------
        Xtarget : (S,)
        """

        # -> (S, T) or (S, T, F)
        X = self._transform_X(X)
        if self.rep == self.REP_M2O:
            if X.ndim == 2:
                X = X[:,:, None]

            # Select last event for T
            X = X[:, -1, dev_slice]
            if X.ndim == 3:
                X = X.squeeze(1)

            assert X.ndim == 2
            return X
        else:
            raise NotImplementedError
    

    def inverse_transform(self, X, y):
        """
        
        
        """
        if self.rep == self.REP_M2M:
            raise NotImplementedError

        elif self.rep == self.REP_M2O:
            raise NotImplementedError



from sklearn.base import BaseEstimator, ClassifierMixin, TransformerMixin
import pandas as pd
import numpy as np
from pyadlml.constants import START_TIME, END_TIME, ACTIVITY, TIME
from pyadlml.dataset._core.activities import add_other_activity, is_activity_df


def _permute(y, start_time, duration=None):
    assert 'other' in y[ACTIVITY].unique()
    y = y.copy().reset_index(drop=True)
    y['diff'] = y[END_TIME] - y[START_TIME]
    df = y.sample(frac=1).reset_index(drop=True)
    df[START_TIME] = df['diff'].shift(1)
    df.at[0, START_TIME] = start_time
    df[START_TIME] = df[START_TIME].cumsum()
    df[START_TIME] = pd.to_datetime(df[START_TIME])
    df[END_TIME] = df[START_TIME] + df['diff']
    df[END_TIME] -= pd.Timedelta('10ns')

    if duration is None:
        assert abs((y[END_TIME] - y[START_TIME]).sum() -
                   (df[END_TIME] - df[START_TIME]).sum()) <= pd.Timedelta('1ms')
    else:
        df['diff'] = df['diff'].cumsum()
        mask = df['diff'] < duration
        mask[mask.idxmin()] = True
        df = df[mask]
        # Cut last activity to size
        df.at[df.index[-1], END_TIME] += df.loc[df.index[-1], 'diff'] - duration

    df = df.drop(columns='diff')
    return df


class RandomPermuteTransformer(TransformerMixin):
    """ Permutates the validation set
    """

    def transform(self, y):
        return _permute(y, y[START_TIME].min())


class RandomPermuteEstimator():
    """ Permutates fractions of the train set as long as the val set
    """

    def __init__(self):
        self.y_train_permuted_ = None

    def fit(self, y_train):
        if 'other' not in y_train[ACTIVITY].unique():
            y_train = add_other_activity(
                y_train, min_diff=pd.Timedelta('100ns'))
        self.y_train_permuted_ = RandomPermuteTransformer().transform(y_train)
        return self

    def predict(self, y):
        dur = y[END_TIME].max() - y[START_TIME].min()
        y_pred = _permute(self.y_train_permuted_, y[START_TIME].min(), dur)
        return y_pred


class RandomSplitTransformer(TransformerMixin):
    """
    """

    def transform(self, y):
        # Make a copy of the input data frame to avoid modifying the original
        df = y.copy()
        df_without_other = y[y[ACTIVITY] != 'other']
        freq = (df_without_other[END_TIME] -
                df_without_other[START_TIME]).min()/10

        # Set the start_time as the index
        df = df.set_index(START_TIME)
        df['activity'] = df['activity'].astype('category')
        # import dask.dataframe as dd
        # from dask.distributed import Client
        # client = Client()  # start distributed scheduler locally.

        # Resample the activity column using the specified frequency
        resampled_activities = df['activity'].resample(
            freq, origin='start').ffill()

        # resampled_activities = dd.from_pandas(resampled_activities, npartitions=10)
        # Shuffle
        resampled_activities = resampled_activities.sample(frac=1)

        # Decrease dataframe size by
        resampled_activities = resampled_activities[resampled_activities != resampled_activities.shift(
        )]
        resampled_activities = resampled_activities.reset_index()
        # crashes at this line
        resampled_activities[END_TIME] = resampled_activities[START_TIME].shift(
            -1)
        # resampled_activities.compute()
        resampled_activities.at[resampled_activities.index[-1],
                                END_TIME] = y.at[y.index[-1], END_TIME]
        resampled_activities[END_TIME] -= pd.Timedelta('10ns')

        return resampled_activities


class RandomSplitEstimator():
    """
    """

    def __init__(self):
        self.y_train_permuted_ = None

    def fit(self, y_train):
        self.y_train_permuted_ = RandomSplitTransformer().transform(y_train)
        return self

    def predict(self, y):
        dur = y[END_TIME].max() - y[START_TIME].min()

        df = self.y_train_permuted_.copy()
        df['diff'] = df[END_TIME] - df[START_TIME]
        df['diff'] = df['diff'].cumsum()
        mask = df['diff'] < dur
        mask[mask.idxmin()] = True
        df = df[mask]
        # Cut last activity to size
        df.at[df.index[-1], END_TIME] += df.loc[df.index[-1], 'diff'] - dur

        return df


class MaxPrevalenceEstimator(BaseEstimator, ClassifierMixin):
    """ Always predict most prominent class
    """

    def fit(self, X, y=None):
        X = X.copy()
        X['duration'] = X[END_TIME] - X[START_TIME]
        self.activity_ = X.groupby(ACTIVITY)['duration'].sum().idxmax()
        return self

    def predict(self, y):
        if is_activity_df(y):
            return pd.DataFrame({
                START_TIME: [y[START_TIME].min()],
                END_TIME: [y[END_TIME].max()],
                ACTIVITY: [self.activity_],
            })
        else:
            return np.array([self.activity_] * len(y))


class MarkovChainSampler(BaseEstimator):
    """ Learns transition matrix and generates random sequences by sampling 
        from seq and then the means
    """

    def __init__(self, dt='s'):
        self.dt = dt

    def plot_activity_dist(self, activity, n_bins=200, file_path=None):
        import matplotlib.pyplot as plt
        import numpy as np
        from scipy.stats import lognorm
        df = self.df_act_
        # Extract the asdf column from the dataframe
        asdf = df[df[ACTIVITY] == activity]['duration']
        z = {v: p for p, v in self.int2act.items()}[activity]
        # Plot the histogram with logarithmic x-axis
        # plt.xscale('log')

        fig, ax = plt.subplots()
        ax.set_xlabel('seconds')
        ax.set_ylabel('Frequency')
        ax.set_title('Histogram other')

        # Plot the probability density function
        hist, bin_edges = np.histogram(asdf, bins=n_bins)
        x = np.linspace(bin_edges[0], bin_edges[-1], 10000)
        pdf, params = self.emissions_[z]

        ax.plot(x, pdf(*params).pdf(x), 'red', linewidth=1)
        hist, bin_edges, _ = plt.hist(
            asdf, bins=n_bins, density=True, alpha=0.5)

        # Calculate the range of the data and add 20% to each end of the x-axis limits
        x_range = bin_edges[-1] - bin_edges[0]
        x_pad = 0.1 * x_range
        ax.set_xlim(bin_edges[0] - x_pad, bin_edges[-1] + x_pad)
        if file_path is None:
            return fig
        else:
            fig.savefig(file_path)

    def plot_activity_dists(self, n_bins=50, file_path=None):
        import matplotlib.pyplot as plt
        import numpy as np
        from scipy.stats import lognorm
        df = self.df_act_

        # Calculate the number of rows and columns for the subplots
        n_rows = len(self.emissions_)
        n_cols = 3
        if n_rows % n_cols == 0:
            n_rows //= n_cols
        else:
            n_rows = (n_rows // n_cols) + 1

        fig, axs = plt.subplots(n_rows, n_cols,  figsize=(12, 4 * n_rows))
        axs = axs.flatten()

        for i, (z, (pdf, params)) in enumerate(zip(self.int2act, self.emissions_)):
            activity = self.int2act[z]
            # Extract the asdf column from the dataframe for the current activity
            asdf = df[df[ACTIVITY] == activity]['duration']

            # Get the current axis for the plot
            ax = axs[i]
            # Plot the histogram with logarithmic x-axis
            # ax.set_xscale('log')
            ax.set_xlabel('seconds')
            ax.set_ylabel('Frequency')
            ax.set_title(activity)

            # Plot the probability density function
            x = np.linspace(-200, 10000, 14000)
            ax.plot(x, pdf(*params).pdf(x), 'red', linewidth=1)
            hist, bin_edges, _ = ax.hist(
                asdf, bins=n_bins, density=True, alpha=0.5)

            # Calculate the range of the data and add 20% to each end of the x-axis limits
            x_range = bin_edges[-1] - bin_edges[0]
            x_pad = 0.2 * x_range
            ax.set_xlim(bin_edges[0] - x_pad, bin_edges[-1] + x_pad)

        # Hide any unused subplots
        for i in range(len(self.emissions_), n_rows * n_cols):
            axs[i].axis('off')
        plt.tight_layout(rect=[0, 0.05, 1, 0.95])

        if file_path is None:
            return fig
        else:
            fig.savefig(file_path)

    def fit(self, y: pd.DataFrame):
        """ 
        Parameters
        ----------
        y : pd.DataFrame
            Activity dataframe

        """
        from pyadlml.stats import activity_duration
        from pyadlml.stats import activity_transition
        from scipy.stats import lognorm, norm
        from pyadlml.dataset.stats.activities import _get_freq_func
        from pyadlml.dataset._core.activities import add_other_activity

        transitions = activity_transition(y)
        self.int2act = {k: v for k, v in enumerate(transitions.index)}
        self.int2act[len(self.int2act)] = 'other'
        self.act2int = {v: k for k, v in self.int2act.items()}

        self.transitions_ = np.divide(
            transitions.values, transitions.values.sum(axis=1)[:, np.newaxis])
        p = activity_duration(y, unit='s', normalize=True)
        self.pi_ = p.sort_index(axis=0)
        assert (transitions.index == self.pi_[ACTIVITY]).all()

        df = add_other_activity(y)
        # integrate the time difference for activities
        diff = 'total_time'
        df[diff] = df[END_TIME] - df[START_TIME]
        df = df[[ACTIVITY, diff]]

        df['duration'] = df[diff].apply(_get_freq_func('s'))
        self.df_act_ = df

        # Fit best length distribution
        self.emissions_ = []
        df['score'] = df[ACTIVITY].apply(self.act2int.get)
        df = df.sort_values(by='score')
        df = df.drop(columns='score')
        grouped_activities = df.groupby('activity', sort=False)['duration']

        from scipy.stats import norm, expon, lognorm, kstest, gaussian_kde
        # Define a list of distributions to try fitting
        distributions = [norm, expon, lognorm]
        for activity, group in grouped_activities:
            if activity in ['visiting_toilet']:
                print('asdf')
            # Fit the data to each distribution and calculate the best fit parameters
            best_fit = None
            best_fit_params = {}
            best_fit_score = -np.inf
            for distribution in distributions:
                # Get the name of the distribution
                name = distribution.name

                # Fit the data to the distribution and get the parameters
                params = distribution.fit(group)
                # Calculate the score of the fit using the Kolmogorov-Smirnov test
                # This measures how well the fitted distribution matches the data
                _, p_value = kstest(group, distribution.cdf, args=params)

                # If the p-value is higher than the previous best, update the best fit
                if p_value > best_fit_score:
                    best_fit = distribution
                    best_fit_params = params
                    best_fit_score = p_value

            self.emissions_.append((best_fit, best_fit_params))

        return self

    def sample(self, y):
        eps = pd.Timedelta('10ns')
        data = []
        C = self.transitions_.shape[0]

        # TODO refactor is count more appropriate????
        z_t = np.random.choice(np.arange(C), p=self.pi_['s'])
        t = y[START_TIME].min()
        while t < y[END_TIME].max():

            # Sample transition
            z_t = np.random.choice(np.arange(C), p=self.transitions_[z_t])

            # Sample z
            pdf, params = self.emissions_[z_t]
            dur_z = max(eps, pd.Timedelta(seconds=pdf.rvs(*params, size=1)[0]))

            # Sample other
            pdf, params = self.emissions_[self.act2int['other']]
            dur_other = max(eps, pd.Timedelta(
                seconds=pdf.rvs(*params, size=1)[0]))

            data.append([t, t+dur_other, 'other'])
            t += dur_other + eps
            data.append([t, t+dur_z, self.int2act[z_t]])
            t += dur_z + eps

        data[-1][1] = y[END_TIME].max()
        df = pd.DataFrame(data=data, columns=[START_TIME, END_TIME, ACTIVITY])

        return df


class TimeAwareMarkovChainSampler(MarkovChainSampler):
    def __init__(self, dt='s', time_dt='2h'):
        super().__init__(dt)
        self.time_dt = time_dt

    def fit(self, y):
        super().fit(y)
        from pyadlml.dataset.stats import activity_transition, activity_count
        from pyadlml.plot import plot_activity_transitions
        from pyadlml.feature_extraction import TimeOfDay
        self.tde_ = TimeOfDay(dt=self.time_dt, inplace=True)
        y = y.rename(columns={START_TIME: TIME})
        y = self.tde_.fit_transform(y)
        y = y.rename(columns={TIME: START_TIME})
        activities = y[ACTIVITY].unique()
        C = len(activities)
        activities.sort()

        self.transitions_ = {}
        self.pi_ = {}
        time_bins = []
        for time_bin, group in y.groupby('time_bins'):
            p = activity_count(
                group[[START_TIME, END_TIME, ACTIVITY]]).set_index(ACTIVITY)
            for act in set(activities) - set(p.index):
                p.loc[act] = 0
            p = p.sort_index(axis=0)
            p['occurrence'] += 1e-10
            self.pi_[time_bin] = np.divide(
                p['occurrence'].values, p['occurrence'].values.sum())

            t = activity_transition(group)
            if t.empty:
                t = pd.DataFrame(index=activities, columns=activities)
                t = t.fillna(0)

            for act in set(activities) - set(t.index):
                t.loc[act] = 0
                t[act] = 0
            t += 1e-10
            t = t.sort_index(axis=0)
            self.transitions_[time_bin] = np.divide(
                t.values, t.values.sum(axis=1)[:, np.newaxis])
            time_bins.append(time_bin)

        for time_bin in set(self.tde_.get_time_bins()) - set(time_bins):
            self.pi_[time_bin] = np.ones(C)/C
            self.transitions_[time_bin] = np.ones((C, C))/C

        self.int2act = {k: v for k, v in enumerate(activities)}
        self.int2act[len(self.int2act)] = 'other'
        self.act2int = {v: k for k, v in self.int2act.items()}

        return self

    def sample(self, y):
        eps = pd.Timedelta('10ns')
        data = []

        t = y[START_TIME].min()
        t_bin = self.tde_.get_time_bin(t)
        C = self.transitions_[t_bin].shape[0]

        z_t = np.random.choice(np.arange(C), p=self.pi_[t_bin])
        while t < y[END_TIME].max():

            # Sample transition
            z_t = np.random.choice(
                np.arange(C), p=self.transitions_[t_bin][z_t])

            # Sample z
            pdf, params = self.emissions_[z_t]
            dur_z = max(eps, pd.Timedelta(seconds=pdf.rvs(*params, size=1)[0]))

            # Sample other
            pdf, params = self.emissions_[self.act2int['other']]
            dur_other = max(eps, pd.Timedelta(
                seconds=pdf.rvs(*params, size=1)[0]))

            data.append([t, t+dur_other, 'other'])
            t += dur_other + eps
            data.append([t, t+dur_z, self.int2act[z_t]])
            t += dur_z + eps
            t_bin = self.tde_.get_time_bin(t)

        data[-1][1] = y[END_TIME].max()
        df = pd.DataFrame(data=data, columns=[START_TIME, END_TIME, ACTIVITY])

        return df



from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import KernelDensity
from sklearn.svm import SVC
from pyadlml.pipeline import Pipeline
from pyadlml.preprocessing import StateVectorEncoder, DropTimeIndex, LabelMatcher
from pyadlml.constants import *
from pyadlml.dataset import load_act_assist
import unittest
import sys
import pathlib
working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))


class TestPipeline(unittest.TestCase):
    def setUp(self):
        dataset_dir = str(script_directory) + '/datasets/partial_dataset'
        self.data = load_act_assist(dataset_dir)
        self.df_acts = self.data['activities']
        self.df_devs = self.data['devices']

        self.rf_pipe = [
            ('raw', StateVectorEncoder(encode='raw')),
            ('lbl', LabelMatcher(idle=True)),
            ('drop_tidx', DropTimeIndex()),
            ('classifier', RandomForestClassifier(random_state=42))
        ]
        self.kde_pipe = [
            ('raw', StateVectorEncoder(encode='raw')),
            ('drop_tidx', DropTimeIndex()),
            ('classifier', KernelDensity(kernel='gaussian', bandwidth=0.5))
        ]
        self.svm_pipe = [
            ('raw', StateVectorEncoder(encode='raw')),
            ('lbl', LabelMatcher(idle=True)),
            ('drop_tidx', DropTimeIndex()),
            ('classifier', SVC())
        ]

    # def test_decision_function(self):
    #    #  Apply transforms, and decision_function of the final estimator
    #    pipe = Pipeline(self.svm_pipe)
    #    pipe = pipe.fit(df_devs,
    #                df_acts)
    #    tmp = pipe.decision_function(df_devs)

    def test_fit(self):
        df_devs, df_acts = self.df_acts.copy(), self.df_devs.copy()
        pipe = Pipeline(self.rf_pipe)
        tmp = pipe.fit(df_devs,
                       df_acts)

    def test_fit_predict(self):
        # Applies fit_predict of last step in pipeline after transforms.

        df_devs, df_acts = self.df_acts.copy(), self.df_devs.copy()
        pipe = Pipeline(self.rf_pipe)
        tmp = pipe.fit_predict(
            df_devs,
            df_acts)
        print(tmp)

    def test_fit_transform(self):
        # Fit the model and transform with the final estimator
        df_devs, df_acts = self.df_acts.copy(), self.df_devs.copy()

        pipe = Pipeline(self.rf_pipe)
        tmp = pipe.fit_transform(
            df_devs,
            df_acts)

    def test_get_params(self):
        # Get parameters for this estimator.
        df_devs, df_acts = self.df_acts.copy(), self.df_devs.copy()

        pipe = Pipeline(self.rf_pipe)
        tmp = pipe.get_params()
        # print(tmp)

    def test_predict(self):
        # Apply transforms to the data, and predict with the final estimator
        df_devs, df_acts = self.df_acts.copy(), self.df_devs.copy()

        pipe = Pipeline(self.rf_pipe).fit(
            df_devs, df_acts)
        tmp = pipe.predict(df_devs)
        print(tmp)

    def test_predict_log_proba(self):
        # Apply transforms, and predict_log_proba of the final estimator
        df_devs, df_acts = self.df_acts.copy(), self.df_devs.copy()

        pipe = Pipeline(self.rf_pipe).fit(
            df_devs, df_acts)
        tmp = pipe.predict_log_proba(df_devs)
        print(tmp)

    def test_predict_proba(self):
        # Apply transforms, and predict_proba of the final estimator
        df_devs, df_acts = self.df_acts.copy(), self.df_devs.copy()

        pipe = Pipeline(self.rf_pipe).fit(
            df_devs, df_acts)
        tmp = pipe.predict_proba(df_devs)
        print(tmp)

    def test_score(self):
        # Apply transforms, and score with the final estimator
        df_devs, df_acts = self.df_acts.copy(), self.df_devs.copy()

        pipe = Pipeline(self.rf_pipe).fit(
            df_devs, df_acts)
        tmp = pipe.score(
            df_devs,
            df_acts)
        print(tmp)

    def test_score_samples(self):
        # Apply transforms, and score_samples of the final estimator.
        df_devs, df_acts = self.df_acts.copy(), self.df_devs.copy()

        pipe = Pipeline(self.kde_pipe).fit(df_devs)
        tmp = pipe.score_samples(df_devs)
        print(tmp)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from pathlib import Path
import hashlib
import pandas as pd
from pyadlml.pipeline import Pipeline
from torch.utils.data import Dataset
import json
import pickle
import gc


def _calc_class_weights(Y_t, classes):
    import torch
    Y_t = Y_t.astype(np.int32)
    count_per_class = np.bincount(Y_t.flatten())
    assert len(classes) == len(count_per_class)
    class_weights_ = torch.tensor(
        count_per_class.sum()/(len(count_per_class)*count_per_class
    ),dtype=torch.float32)

    return class_weights_

class TorchDataset(Dataset):
    def __init__(self, X, y, transforms: Pipeline, transform_params={}, class_encoding=True, use_cached=False, use_memcache=False):
        import torch

        rand_str = self._get_rand_rep(transforms)
        sub_folder = 'train' if transforms.is_in_train_mode() else 'eval'

        #self.fp_tmp = Path(f'/mnt/ssd256/tmp/{rand_str}/{sub_folder}/')
        self.fp_tmp = Path(f"/home/chris/master_thesis/trainings_cache/{rand_str}/{sub_folder}/")
        #self.hdf5 = hdf5
        if use_memcache and self.fp_tmp.exists():
            with open(self.fp_tmp.joinpath('data.pt'), 'rb') as f:
                tmp_dict = pickle.load(f)
            self.classes_ = tmp_dict['classes']
            self.class_weights_ = tmp_dict['class_weights']
            self.shape = tmp_dict['shape']

        elif use_cached and self.fp_tmp.exists(): 
            with open(self.fp_tmp.joinpath('data.pt'), 'rb') as f:
                tmp_dict = pickle.load(f)

            self.classes_ = tmp_dict['classes']
            self.class_weights_ = tmp_dict['class_weights']
            self.nr_steps = tmp_dict['step']
            self.chunk_size = tmp_dict['chunk_size']
            self.shape = tmp_dict['shape']

            self.int2class_ = {k:v for k, v in enumerate(self.classes_)}
            self.int2class_func = np.vectorize(self.int2class_.get)
            self.class2int_ = {v:k for k, v in self.int2class_.items()}
            self.class2int_func = np.vectorize(self.class2int_.get)
        else:
            if transforms.is_in_train_mode():
                X_t, Y_t = transforms.fit_transform(X.copy(), y.copy(), **transform_params)
            else:
                X_t, Y_t = transforms.transform(X.copy(), y.copy(), **transform_params)

            # Cast X_t and Y_t to numpy
            if isinstance(X_t, pd.DataFrame) or isinstance(X_t, pd.Series):
                X_t = X_t.values
            if isinstance(Y_t, pd.DataFrame) or isinstance(Y_t, pd.Series):
                Y_t = Y_t.values
            assert isinstance(X_t, np.ndarray) and isinstance(Y_t, np.ndarray)

            # Create Y_t class encoding
            self.classes_ = np.unique(Y_t.flatten())
    
            if class_encoding:
                self.int2class_ = {k:v for k, v in enumerate(self.classes_)}
                self.int2class_func = np.vectorize(self.int2class_.get)
                self.class2int_ = {v:k for k, v in self.int2class_.items()}
                self.class2int_func = np.vectorize(self.class2int_.get)
                Y_t = self.class2int_func(Y_t)

            # Compute class frequencies
            # Position 0 equal freq of class with idx zero
            self.class_weights_ = _calc_class_weights(Y_t, self.classes_)

            # Create torch tensor dataset and let torch infer datatype
            # What could go wrong
            import psutil
            self.shape = X_t.shape
            self.N = X_t.shape[0]
            mem_frac = 0.8
            #if hdf5:
            #    self.fp_tmp.mkdir(exist_ok=True, parents=True)
            #    self.fp_tmp = Path(f'/mnt/ssd256/tmp/{rand_str}/{sub_folder}.hdf5')
            #    with h5py.File(self.fp_tmp, "w") as f:
            #        f.create_dataset("X_t", shape=X_t.shape, dtype=np.float32)
            #        f.create_dataset("y_t", shape=Y_t.shape, dtype=np.int64)
            #        f.attrs['classes'] = self.classes_
            #        f.attrs['shape'] = self.shape
            #        f.attrs['class_weights'] = self.class_weights_
            if use_memcache:
                tmp_dict = {}
                tmp_dict['classes'] = self.classes_
                tmp_dict['step'] = self.nr_steps
                tmp_dict['chunk_size'] = self.chunk_size
                tmp_dict['shape'] = self.shape
                tmp_dict['class_weights'] = self.class_weights_
                with open(self.fp_tmp.joinpath('data.pt'), 'wb') as f:
                    pickle.dump(tmp_dict, f)

            if X_t.nbytes > psutil.virtual_memory().available*mem_frac or use_cached:
                self.fp_tmp.mkdir(exist_ok=True, parents=True)
                self.chunk_size = 100
                self.nr_steps = self.N//self.chunk_size
                for c in range(self.nr_steps+1):
                    torch.save(
                        torch.tensor(X_t[self.chunk_size*c:self.chunk_size*c+self.chunk_size]).to(torch.float32),
                        self.fp_tmp.joinpath(f'Xt_chunk_{c}.pt')
                    )
                    torch.save(
                        torch.tensor(Y_t[self.chunk_size*c:self.chunk_size*c+self.chunk_size]).long(), 
                        self.fp_tmp.joinpath(f'Yt_chunk_{c}.pt')
                    )

                # 
                tmp_dict = {}
                tmp_dict['classes'] = self.classes_
                tmp_dict['step'] = self.nr_steps
                tmp_dict['chunk_size'] = self.chunk_size
                tmp_dict['shape'] = self.shape
                tmp_dict['class_weights'] = self.class_weights_
                with open(self.fp_tmp.joinpath('data.pt'), 'wb') as f:
                    pickle.dump(tmp_dict, f)
            else:
                self.Xtr = torch.tensor(X_t).to(torch.float32)
                self.Ytr = torch.tensor(Y_t).long()
        
        gc.collect()
            

    def __len__(self):
        return self.shape[0]

    def _get_rand_rep(self, pipe):
        import hashlib
        return hashlib.md5(str(pipe).encode('utf-8')).hexdigest()
        #rand_str = "".join([str(s)for s in np.random.choice(10,10)])

    @property
    def class_weights(self):
        return self.class_weights_

    @property 
    def dct_class_weights(self):
        return {k:v for k, v in zip(self.class2int_,self.classes_)}


    @property
    def n_classes(self):
        return len(self.classes_)

    def int2class(self, i: int):
        return self.int2class_[i]

    def class2int(self, cls: str):
        return self.class2int_[cls]

    def __getitem__(self, idx):

        import torch
        if hasattr(self, 'chunk_size'):
            c = idx//self.chunk_size
            try:
                Xtr = torch.load(self.fp_tmp.joinpath(f'Xt_chunk_{c}.pt'))
                Ytr = torch.load(self.fp_tmp.joinpath(f'Yt_chunk_{c}.pt'))
            except FileNotFoundError:
                print()
            new_idx = idx - c*self.chunk_size
            return Xtr[new_idx], Ytr[new_idx]
        else:
            return self.Xtr[idx], self.Ytr[idx]

class TorchDataset2(Dataset):
    def __init__(self, X_t, y_t, classes):
        self.Xtr = X_t
        self.Ytr = y_t
        self.class_weights = _calc_class_weights(y_t, classes)

    def __len__(self):
        return len(self.Xtr)

    def __getitem__(self, idx):
        import torch
        X = torch.from_numpy(self.Xtr[idx].copy()).float()
        if not isinstance(self.Ytr[idx], np.ndarray): 
            y_np = np.array(self.Ytr[idx])
        else:
            y_np = self.Ytr[idx]
        y = torch.from_numpy(y_np).long()
        return X, y.squeeze()



class TorchMemcacheDataset(Dataset):
    def __init__(self, pipe: Pipeline, X_t=None, y_t=None):
        import torch

        rand_str = self._get_rand_rep(pipe)
        sub_folder = 'train' if pipe.is_in_train_mode() else 'eval'
        self.fp_tmp = Path(f"/home/chris/master_thesis/trainings_cache/{rand_str}/{sub_folder}/")

        fp_x = self.fp_tmp.joinpath('x.dat')
        fp_y = self.fp_tmp.joinpath('y.dat')

        if self.fp_tmp.exists():
            with open(self.fp_tmp.joinpath('data.pt'), 'rb') as f:
                tmp_dict = pickle.load(f)
            self.class_weights_ = tmp_dict['class_weights']
            self.shape = tmp_dict['X_shape']

            self.Xtr = np.memmap(fp_x, mode='r+', shape=tmp_dict['X_shape'], dtype=tmp_dict['X_dtype'])
            self.Ytr = np.memmap(fp_y, mode='r+', shape=tmp_dict['y_shape'], dtype=tmp_dict['y_dtype'])
        else:

            # Cast X_t and Y_t to numpy
            if isinstance(X_t, pd.DataFrame) or isinstance(X_t, pd.Series):
                X_t = X_t.values
            if isinstance(y_t, pd.DataFrame) or isinstance(y_t, pd.Series):
                y_t = y_t.values
            assert isinstance(X_t, np.ndarray) and isinstance(y_t, np.ndarray)

   
            # Compute class frequencies
            # Position 0 equal freq of class with idx zero
            count_per_class = np.bincount(y_t.flatten())
            self.class_weights_ = torch.tensor(
                count_per_class.sum()/(len(count_per_class)*count_per_class
            ),dtype=torch.float32)

            self.shape = X_t.shape
            self.N = X_t.shape[0]

            if len(y_t.shape) == 1:
                y_t = y_t[:, None]

            tmp_dict = {}
            tmp_dict['class_weights'] = self.class_weights_
            tmp_dict['X_shape'] = X_t.shape
            #tmp_dict['X_dtype'] = X_t.dtype
            tmp_dict['X_dtype'] = np.bool_
            tmp_dict['y_shape'] = y_t.shape
            tmp_dict['y_dtype'] = np.uint8
            #tmp_dict['y_dtype'] = y_t.dtype

            self.fp_tmp.mkdir(parents=True, exist_ok=True)
            with open(self.fp_tmp.joinpath('data.pt'), 'wb') as f:
                pickle.dump(tmp_dict, f)
            self.Xtr = np.memmap(fp_x, mode='w+', shape=tmp_dict['X_shape'], dtype=tmp_dict['X_dtype'])
            self.Ytr = np.memmap(fp_y, mode='w+', shape=tmp_dict['y_shape'], dtype=tmp_dict['y_dtype'])

            self.Xtr[:] = X_t[:]
            self.Ytr[:] = y_t[:]
        
        gc.collect()
            

    def __len__(self):
        return self.shape[0]

    @classmethod
    def _get_rand_rep(cls, pipe):
        return hashlib.md5(str(pipe).encode('utf-8')).hexdigest()

    @property
    def class_weights(self):
        return self.class_weights_

    def __getitem__(self, idx):
        import torch
        X = torch.from_numpy(self.Xtr[idx]).float()
        y = torch.from_numpy(self.Ytr[idx]).long()
        return X, y.squeeze()

    @classmethod 
    def cached_exists(cls, pipe):
        rand_str = cls._get_rand_rep(pipe)
        sub_folder = 'train' if pipe.is_in_train_mode() else 'eval'
        fp_tmp = Path(f"/home/chris/master_thesis/trainings_cache/{rand_str}/{sub_folder}/")

        fp_x = fp_tmp.joinpath('x.dat')
        fp_y = fp_tmp.joinpath('y.dat')
        return fp_x.exists() and fp_y.exists()

       


import unittest
import sys
import pathlib
working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))


class TestImport(unittest.TestCase):
    def setUp(self):
        import pyadlml

    def test_import_datasets(self):
        from pyadlml.dataset import load_act_assist
        from pyadlml.dataset import fetch_amsterdam
        from pyadlml.dataset import fetch_aras
        from pyadlml.dataset import fetch_casas
        from pyadlml.dataset import fetch_mitlab
        from pyadlml.dataset import fetch_tuebingen_2019
        from pyadlml.dataset import fetch_uci_adl_binary

    def test_import_plots(self):
        # import activity plots
        from pyadlml.plot import plot_activities_and_events, plot_activity_boxplot, plot_activity_correction, \
            plot_activity_density, plot_activity_count, plot_activity_transitions

        # import device plots
        from pyadlml.plot import plot_device_event_correlogram, plot_device_event_count, plot_device_event_density, \
            plot_device_event_raster, plot_device_inter_event_times, plot_device_state_boxplot, plot_device_state_cross_correlation, \
            plot_device_state_fractions, plot_device_states

    def test_import_stats(self):
        from pyadlml.stats import activity_dist, activity_count, activity_duration, activity_duration_dist, \
            activity_transition, activity_coverage

        from pyadlml.stats import device_event_count, device_event_cross_correlogram, device_events_one_day, \
            device_firing_rate, device_inter_event_times, device_state_fraction, device_state_similarity, \
            device_state_time

        from pyadlml.stats import contingency_table_events


if __name__ == '__main__':
    unittest.main()



"""
this file is to bring datasets into specific representations
"""
import pandas as pd
import numpy as np
from pyadlml.util import get_parallel, get_npartitions
from pyadlml.constants import OTHER, OTHER_MIN_DIFF, TIME, ACTIVITY, START_TIME, END_TIME


def label_data2(df_devs: pd.DataFrame, df_acts: pd.DataFrame, other=False):
    """



    """
    df = df_acts.copy()
    df_devs = df_devs.copy()



    # A device is counted as belonging to an activity if the timestamp
    # falls into the open interval df_acts[START_TIME] <= ts <= df_acts[END_TIME]
    # For events occuring exactly at the time an activity starts or ends
    # move the activity boundary a bit in the future or past to facilitate 
    # the expected behavior
    assert not df_devs[TIME].duplicated().any() 
    eps = pd.Timedelta('1ns')
    eps_other = '1ns'

    mask_act_et_clsn = df[END_TIME].isin(df_devs[TIME])
    mask_act_st_clsn = df[START_TIME].isin(df_devs[TIME])
    df.loc[df[mask_act_et_clsn].index, END_TIME] += eps
    df.loc[df[mask_act_st_clsn].index, START_TIME] -= eps


    df = df.rename(columns={START_TIME: TIME})

    # Fill up with other
    df_other = df[[END_TIME, ACTIVITY]]
    df_other.loc[:, ACTIVITY] = OTHER
    df_other = df_other.rename(columns={END_TIME: TIME})
    df = df.drop(columns=END_TIME)
    df = pd.concat([df, df_other.iloc[:-1]], ignore_index=True, axis=0) \
        .sort_values(by=TIME) \
        .reset_index(drop=True)
    df['diff'] = df[TIME].shift(-1) - df[TIME]
    mask_invalid_others = (df['diff'] < eps_other) & (
        df[ACTIVITY] == OTHER)
    df = df[~mask_invalid_others][[TIME, ACTIVITY]]

    # CAVE replace the ending 
    df = pd.concat([df, pd.Series({
        TIME: df_acts.at[df_acts.index[-1], END_TIME],
        ACTIVITY: OTHER}
    ).to_frame().T]).reset_index(drop=True)

    df_label = df.copy()
    df_label = df_label.sort_values(by=TIME).reset_index(drop=True)
    df_devs = df_devs.sort_values(by=TIME).reset_index(drop=True)


    # if prediction is larger than ground truth activities pad gt with 'other'
    if df_devs[TIME].iat[0] < df_label[TIME].iat[0]:
        df_label = pd.concat([df_label,
            pd.Series({
                TIME: df_devs.at[0, TIME] - pd.Timedelta('1ms'), 
                ACTIVITY: OTHER
            }).to_frame().T], axis=0, ignore_index=True)

    df_dev_tmp = df_devs.copy()
    df_act_tmp = df_label.copy().sort_values(by=TIME).reset_index(drop=True)

    # Add columns
    df_dev_tmp[ACTIVITY] = np.nan
    for col in set(df_devs.columns).difference(df_act_tmp.columns):
        df_act_tmp[col] = np.nan

    df = pd.concat([df_devs, df_act_tmp], ignore_index=True, axis=0)\
        .sort_values(by=TIME)\
        .reset_index(drop=True)

    df[ACTIVITY] = df[ACTIVITY].ffill()
    df = df.dropna().reset_index(drop=True)
    assert len(df) == len(df_devs)

    if not other:
        other_mask = (df[ACTIVITY] == OTHER)
        df[other_mask] = np.nan

    return df


def label_data(df_devs: pd.DataFrame, df_acts: pd.DataFrame, other=False, n_jobs=1, inplace=True):
    """
    Label a dataframe with corresponding activities based on a time-index.

    Parameters
    ----------
    df_devs : pd.DataFrame
        some data representation that possesses a column 'time' including timestamps.
    df_acts : pd.DataFrame
        a datasets activities. TODO
    other : bool, optional, default=False
        if true this leads to datapoints not falling into a logged activity to be
        labeled as "other"
    n_jobs : int, optional, default=1
        the number of jobs that are run in parallel TODO look up sklearn
    inplace : bool, optional, default=True
        determines whether a new column is appended to the existing dataframe.

    Examples
    --------
    >>> raw = DiscreteEncoder()
    >>> raw
    1 time                    0   ...      13
    2 2008-03-20 00:34:38  False  ...    True
    3 2008-03-20 00:34:39  False  ...   False

    now include
    >>> label_data(raw, data.df_activities, other=True, n_jobs=10)
    1 time                    0   ...      13 activity
    2 2008-03-20 00:34:38  False  ...    True other
    3 2008-03-20 00:34:39  False  ...   False act1

    Returns
    -------
    df : pd.DataFrame
    """

    df_devs = df_devs.copy()
    df_devs[ACTIVITY] = -1

    if n_jobs == 1:
        df_devs[ACTIVITY] = df_devs[TIME].apply(
            _map_timestamp2activity,
            df_act=df_acts,
            other=other)
    else:
        N = get_npartitions()
        if n_jobs == -1 or n_jobs > N:
            n_jobs = N

        import dask.dataframe as dd
        # ddf_activities = dd.from_pandas(df_activities, npartitions=get_npartitions())
        # compute with dask in parallel
        df_devs[ACTIVITY] = dd.from_pandas(df_devs[TIME], npartitions=n_jobs).\
            map_partitions(  # apply lambda functions on each partition
            lambda df: df.apply(
                _map_timestamp2activity,
                df_act=df_acts,
                other=other)).\
            compute(scheduler='processes')
    return df_devs


def _map_timestamp2activity(timestamp, df_act, other):
    """ Map the given timestamp map to an activity in df_act

    Parameters
    ----------
    timestamp : pd.Timestamp
        E.g timestamp 2008-02-26 00:39:25
    df_act : pd.DataFrame
        An activity dataframe
    other : boolean
        Whether to map gaps to NAT or "other"

    Returns
    -------
        label of the activity
    """

    # select activity intervals that the timestamp falls into
    mask = (df_act[START_TIME] <= timestamp) & (timestamp <= df_act[END_TIME])
    matches = df_act[mask]
    match_amount = len(matches.index)

    # 1. case no activity interval matched
    if match_amount == 0 and other:
        return OTHER
    elif match_amount == 0 and not other:
        return pd.NaT

    # 2. case single row matches
    elif match_amount == 1:
        return matches.activity.values[0]

    # 3. case multiple rows
    else:
        print()
        print('*'*70)
        print('ts: ', timestamp)
        print('matches: ', matches)
        print('matches_amount: ', match_amount)
        print('overlap of activities. this should be handled when loading activities')
        print('*'*70)
        raise ValueError



import numpy as np
import pandas as pd
from pyadlml.constants import *
from pyadlml.dataset import load_act_assist
import unittest
import sys
import pathlib

from pyadlml.dataset._core.activities import add_other_activity
working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))


class TestMetrics(unittest.TestCase):
    def setUp(self):
        dataset_dir = str(script_directory) + '/datasets/debug_dataset'
        self.data = load_act_assist(dataset_dir)
        self.df_acts = self.data['activities']['chris']
        self.df_devs = self.data['devices']

    def _gen_identical(self, df_acts, other=False):
        df = add_other_activity(df_acts) if other else df_acts
        y_true = df[ACTIVITY].to_numpy()
        y_true_times = df[START_TIME].to_numpy()
        return y_true, y_true_times

    def _gen(df_acts, self):
        y_pred = df_acts[ACTIVITY]
        y_times = np.array([pd.Timestamp(ts) for ts in ['']])

    def test_online_acc(self):
        df_devs, df_acts = self.df_devs.copy(), self.df_acts.copy()
        activities = np.append(df_acts[ACTIVITY].unique(), 'other')
        from pyadlml.metrics import online_accuracy

        # Check if identity is one
        y_true, y_true_times = self._gen_identical(df_acts, other=True)
        # acc_micro = online_accuracy(df_acts, y_true, y_true_times,
        #                            len(activities), average='micro')
        # assert np.isclose(acc_micro, 1.0, atol=0.001)
        acc_macro = online_accuracy(df_acts, y_true, y_true_times,
                                    len(activities), average='macro')
        assert np.isclose(acc_macro, 1.0, atol=0.001)

        # Check if without other is one
        y_pred, y_pred_times = self._gen_identical(df_acts, other=False)
        acc_micro = online_accuracy(df_acts, y_pred, y_pred_times,
                                    len(activities), average='micro')
        acc_true = 0.874
        assert np.isclose(acc_micro, acc_true, atol=0.001), \
            f'Should have {acc_true} but got {acc_micro}'

        acc_macro = online_accuracy(df_acts, y_pred, y_pred_times,
                                    len(activities), average='macro')
        assert np.isclose(acc_macro, acc_true, atol=0.001), \
            f'Should have {acc_true} but got {acc_macro}'

    def test_online_ppv(self):
        df_devs, df_acts = self.df_devs.copy(), self.df_acts.copy()

    def test_online_tpr(self):
        df_devs, df_acts = self.df_devs.copy(), self.df_acts.copy()

    def test_online_conf_mat(self):
        df_devs, df_acts = self.df_devs.copy(), self.df_acts.copy()


if __name__ == '__main__':
    unittest.main()



from pyadlml.constants import *
from pyadlml.dataset import load_act_assist
import unittest
import sys
import pathlib

from pyadlml.feature_extraction import DayOfWeek, extract_time_difference
working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))


class TestDirtyDataset(unittest.TestCase):
    def setUp(self):
        dataset_dir = script_directory.joinpath('/datasets/dirty_dataset')
        self.data = load_act_assist(dataset_dir)
        self.df_acts = self.data['activities']
        self.df_devs = self.data['devices']

    def test_extract_time_bins(self):
        df_devs = self.df_devs.copy()
        df_res2h = extract_time_difference(df_devs)
        df_res6h = extract_time_difference(df_devs, resolution='6h')
        df_res10m = extract_time_difference(df_devs, resolution='10m')
        df_res40m = extract_time_difference(df_devs, resolution='40m')
        df_res10s = extract_time_difference(df_devs, resolution='10s')

        assert len(df_res2h.columns) == 12
        assert len(df_res6h.columns) == 4
        assert len(df_res10m.columns) == 144
        assert len(df_res40m.columns) == 36
        assert len(df_res10s.columns) == 8640

    def test_extract_day_of_week(self):
        df_devs = self.df_devs.copy()
        df_dow = DayOfWeek(one_hot_encoding=False)


if __name__ == '__main__':
    unittest.main()



from torch import nn
from pyadlml.model.nn import Tanh, Embedding, FlattenConsecutive
from torch.nn import Linear, BatchNorm1d
import torch 
import torch.nn.functional as F
import numpy as np
from pyadlml.model.nn.util import CausalConv1d, DilatedConv1d
import math

def print_in_out(layer, name=None):
    def wrapper(x):
        i = list(x.shape)
        x = layer(x)
        j = list(x.shape)
        name = layer.__class__.__name__ if name is None else name
        print(f'{name} {i} -> {j}')
        return x
    return wrapper


class DilatedModel(nn.Module):

    def __init__(self, n_features, n_classes, seq_length, n_emb=3, n_hidden=100, dilation_factor=2):
        """
        """
        super().__init__()

        # Calculate 
        nr_layers = math.log(seq_length, dilation_factor)
        assert nr_layers % 1 == 0, 'Sequence length has to be a exponential of dilation factor to form usable layers'

        class CausalBlock(nn.Module):
            def __init__(self, dilation_factor, lin_in_dim, n_hidden):
                super().__init__()
                self.d = dilation_factor
                self.flc = FlattenConsecutive(dilation_factor)      # B, T, E     -> B, T/d, E*d
                self.lin = Linear(lin_in_dim, n_hidden, bias=False) # B, T/d, E*d -> B, T/d, H
                self.bn1d = BatchNorm1d(n_hidden)                   # B, T/d, H   -> B, T/d, H
                self.tanh = Tanh()                                  # B, T/d, H   -> B, T/d, H

            def forward(self, x: torch.Tensor) -> torch.Tensor:
                B, T, C = x.shape

                x = self.flc(x)
                T = T//self.d

                x = self.lin(x)
                C = self.lin.out_features

                x = x.view(B, C, T)
                x = self.bn1d(x)
                x = x.view(B, T, C)
                self.tanh(x)
                return x

        layers = []
        for i in range(int(nr_layers)):
            if i == 0:
                in_lin_dim = n_features*dilation_factor
            else:
                in_lin_dim = n_hidden*dilation_factor
            layers.append(CausalBlock(dilation_factor, in_lin_dim, n_hidden))

        self.layers = nn.ModuleList(layers)
        self.ff = nn.Linear(n_hidden, n_classes)


    def compute_loss(self, logits_cls, y_cls, weight):
        loss_cls = F.cross_entropy(logits_cls, y_cls, weight=weight) 
        # Refactor
        praefix = 'train' if self.training else 'val'

        return loss_cls

    @classmethod
    def gen_nr_layer_dil_pairs(cls, dilation_factor, nr_lengths=10):
        return [(i, dilation_factor**i) for i in range(1, nr_lengths+1)]


    def forward(self, x):
        """
        """
        B, T, F = x.shape

        # (B,T,F) -> ()
        for l in self.layers:
            x = l(x)

        x = self.ff(x)

        return x.squeeze()


class ResidualBlock(torch.nn.Module):
    def __init__(self, res_channels, skip_channels, dilation):
        """
        Residual block
        :param res_channels: number of residual channel for input, output
        :param skip_channels: number of skip channel for output
        :param dilation:
        """
        super(ResidualBlock, self).__init__()

        self.dilate_filter = DilatedConv1d(
            res_channels, res_channels,
            kernel_size=2, stride=1,  
            dilation=dilation,
            bias=False
        )  
        
        self.dilate_gate = DilatedConv1d(
            res_channels, res_channels,
            kernel_size=2, stride=1,  
            dilation=dilation,
            bias=False
        )

        # 1x1 convolutions
        self.conv_res = torch.nn.Conv1d(res_channels, res_channels, 1)
        self.conv_skip = torch.nn.Conv1d(res_channels, skip_channels, 1)

    def forward(self, x, skip_size):
        """
        Parameters
        ----------
        x : torch.tensor of shape (B, R, T) 
            where B is batch size, R is #residual dimensions and T is the sequence length

        Returns
        -------
        torch.tensor of shape (B, S, T)
            where B is batch size, S is the #skip dimensions and T is the sequence length

        """
        B, R, T = x.shape

        # Follows eq. 2 from paper rather than diagram
        x_sigm = self.dilate_filter(x)  # (B, R, T) -> (B, R, T)
        x_tanh = self.dilate_gate(x)    # (B, R, T) -> (B, R, T)

        # PixelCNN gate
        gated_tanh = F.tanh(x_tanh)
        gated_sigmoid = F.sigmoid(x_sigm)
        gated = gated_tanh * gated_sigmoid

        # Residual connection
        output = self.conv_res(gated)

        x = x + output

        # Skip connection
        skip = self.conv_skip(gated)

        return output, skip


class WaveNetModel(torch.nn.Module):
    """
    Inspired by:
        https://github.com/golbin/WaveNet/blob/master/wavenet/networks.py 
    and
        https://github.com/vincentherrmann/pytorch-wavenet/blob/26ba28989edcf8688f6216057aafda07601ff07e/wavenet_model.py#L144

    
    """
    def __init__(self, n_layers, n_blocks, in_channels, res_channels, skip_channels, 
                       end_channels, n_features, n_classes):
        """
        Stack residual blocks by layer and stack size

        Parameters
        ----------
        :param layer_size: integer, 10 = layer[dilation=1, dilation=2, 4, 8, 16, 32, 64, 128, 256, 512]
        :param stack_size: integer, 5 = stack[layer1, layer2, layer3, layer4, layer5]
        :param in_channels: number of channels for input data. skip channel is same as input channel
        :param res_channels: number of residual channel for input, output
        :return:
        """
        super(WaveNetModel, self).__init__()
        self.layer_size = n_layers
        self.stack_size = n_blocks
        self.res_channels = res_channels
        self.skip_channels = skip_channels
        self.in_channels = in_channels

        self.receptive_fields =  self.calc_receptive_fields(n_layers, n_blocks)

        # Build causal conv
        # Padding=1 for same size(length) between input and output for causal convolution
        self.causal = CausalConv1d(in_channels, res_channels, kernel_size=2, stride=1, bias=False)

        # Build Residual layers
        # For s=3, l=4 -> [1,2,4,8,  1,2,4,8,  1,2,4,8]
        dilations = [ 2**l for s in range(0, self.stack_size) for l in range(0, self.layer_size)]
        
        self.res_layers = torch.nn.ModuleList([
                            ResidualBlock(res_channels, skip_channels, dilation)
                            for dilation in dilations
        ])

        self.densnet = torch.nn.Sequential(*[
            torch.nn.ReLU(),
            torch.nn.Conv1d(skip_channels, end_channels, kernel_size=1),
            torch.nn.ReLU(),
            torch.nn.Conv1d(end_channels, n_features, kernel_size=1),
            torch.nn.Softmax(dim=1),
        ])

        # No downsampling for low timestep sizes
        self.classif_net = torch.nn.Sequential(*[
            torch.nn.ReLU(),
            # (B,S,T) -> (B, 100, T)
            torch.nn.Conv1d(skip_channels, 100, kernel_size=5),
            torch.nn.ReLU(),
            torch.nn.Conv1d(100, 50, kernel_size=3),
            torch.nn.ReLU(),
            torch.nn.Conv1d(50, 20, kernel_size=1),
            torch.nn.ReLU(),
            torch.nn.Flatten(),
            torch.nn.Linear(1880, n_classes)
            #torch.nn.Linear(20*10, n_classes)
        ])

    @staticmethod
    def calc_receptive_fields(layer_size, stack_size):
        
        layers = [2 ** i for i in range(0, layer_size)] * stack_size
        num_receptive_fields = np.sum(layers)

        return int(num_receptive_fields)

    def calc_output_size(self, x):
        output_size = int(x.size(2)) - self.receptive_fields

        self.check_input_size(x, output_size)

    def forward(self, x):
        """
        The size of timestep(3rd dimention) has to be bigger than receptive fields
        :param x: Tensor[batch, timestep, channels]
        :return: Tensor[batch, timestep, channels]

        Parameters
        ----------
        x: torch.tensor
            tensor with shape (B, T, C)

        Returns
        -------
        torch.tensor
            Shape
        """
        B, T, C, R, S = [*x.shape, self.res_channels, self.skip_channels]

        # (B,T,C) -> (B,C,T)
        # Move T to last dimension since most torch operations assume seq length as such
        output = x.transpose(1, 2)   

        # (B,C,T) -> (B,R,T)
        output = self.causal(output)

        skip_stream = torch.zeros((B, S, T), dtype=output.dtype, device=output.device)
        for res_block in self.res_layers:
            # (B,R,T) -> (B,R,T), (B,S,T)
            output, skip_i = res_block(output, S)
            skip_stream = skip_stream + skip_i

        # (B,S,T) -> (B,C,T)
        output = self.densnet(skip_stream)

        logits_next_dev = output.swapaxes(1,2)

        # Take last ouput after applying all res_blocks
        logits_cls = self.classif_net(skip_i)


        return logits_next_dev, logits_cls


    def compute_loss(self, logits_next_dev, y_next_dev, logits_cls, y_cls, weight, rep='many-to-many'):
        B, T, C_d, C_c = [*logits_next_dev.shape, y_cls.shape[-1]]
        # TODO critical, build data loader for next item prediction
        if rep == 'many-to-one':
            # Take last symbol as next dev prediction many-to-one 
            # (B,T,C_d) -> (B,C_d)
            logits_next_dev = logits_next_dev[:,:-1,:]
            # (B,T) -> (B,)
            y_next_dev = y_next_dev[:, -1]
            
        if rep == 'many-to-many':
            # (B,T,C_d) -> (B*T, C_d)
            logits_next_dev = logits_next_dev.reshape(B*T, C_d)
            # (B,T) -> (B*T)
            y_next_dev = y_next_dev.view(B*T)

        

        loss_cls = F.cross_entropy(logits_cls, y_cls, weight=weight) 
        loss_nd = F.cross_entropy(logits_next_dev, y_next_dev)

        loss = loss_cls + loss_nd

        # Refactor
        praefix = 'train' if self.training else 'val'
        import mlflow
        mlflow.log_metric(f'{praefix}_loss_cls', loss_cls.item())
        mlflow.log_metric(f'{praefix}_loss_next_dev', loss_nd.item())

        return loss
